/* alpha SRAM definition

INPUT BITS

	0-5		left input
	6-11	right input
	12-15	a0..a4 operation
	12-16	a5 operation

OUTPUT BITS

	0-5		result
	6		carry
	7		propagate
	8-12	a5 encoded ranges / ASL / ASR

RANGE ENCODING

	Sum of two alternatives based on carry decision 5 clear or set:

	clr	set		result
	------------------------------
	 0	 0		too negative
	 1	 5		fits signed only
	 2	10		fits always
	 3	15		fits unsigned only
	 4	20		too positive

	Above applies to fs_add_u, fs_add_s, fs_cmp, fs_bound.

ASL AND ASR ENCODING

	Sum of:

	1 (bit 8)	L35 is set
	2 (bit 9)	R = 0

	C	for ASR, 0 encroaches bit 35 (e) or 36 (E)
	P	for ASR, 1 encroaches bit 35 (e) or 36 (E)

MH ENCODING

	0				no overflow
	3 (bits 8 & 9)	short mul exceeds 36 bits

*/

#include "vm.h"

/* alpha5 is double-size, leading to odd spec in C */
static uint16_t a0[1 << 16], a1[1 << 16], a2[1 << 16],
	a3[1 << 16], a4[1 << 16], a5[2 << 16];
uint16_t *alpha[] = { a0, a1, a2, a3, a4, a5 };

// Consider high tribble of a word for range encoding.
// If d5 is set, result must be multiplied by 5.
int encode_range(int q)
{
	if (q < -32) return 0;		// too  negative
	if (q <   0) return 1;		// fits signed only
	if (q <  32) return 2;		// fits always
	if (q <  64) return 3;		// fits unsigned only
	return 4;					// too  positive
}

/* compute a single value in the "regular" (slots 0-15) alpha table */
int alpha_operation(int slot, int subword, int l, int r)
{
	int al = 0, t, lim = 35, bit;
	int name = alpha_slots[slot];

	if (l<0 || r<0 || l>c63 || r > c63 || subword < 0 || subword > c5)
		oops("alpha_operation");

	/* see "alpha operations" table from PDF */
	switch (name) {
		case '=':

			// For a0-a4, do a magnitude comparison.
			// a5's magnitude comparison is in its high-memory section.
			al = l;
			if (subword < 5) {
				if (l < r)
					al |= a_carry_bit;
				else if (l > r)
					al |= a_prop_bit;
			}

A5:			if (subword == 5) {
				if (l & bit5) {
					al |= a_carry_bit;		// theta: ASR.su/ss sign extension
N:					al |= bit8;				// flags: L35 was set for ASR
				}
				if (!r)
					al |= bit9;				// flags: R == 0 for ASR
			}
			break;

		case 'm':
			l ^= r, r ^= l, l ^= r;
			nobreak;

		case '-':
			r = c63 & ~r;
			nobreak;

		case '+':
			al = l + r;
			if (al == c63)
				al |= a_prop_bit;
			break;

		case '&':
			al = l & r;
			break;

		case '|':
			al = l | r;
			break;

		case 'x':
			al = c63 & l * r;
			break;

		case 'X':
			al = c63 & l * r >> c6;
			if (al && subword == 5)
				al |= bit8 | bit9;		// fs_fsr_us works for this
			break;

		case '*':						// DSL only
			al = subword == c5 ? r : l;
			break;

		case 'E':						// left shift bit 36 encroached?
			lim = 36;
			nobreak;

		case 'e':						// left shift bit 35/36 encroached?
			al = l;
			for (bit=0; bit<6; ++bit) {
				if (r < lim - 6 * subword - bit)
					continue;			// no encroachment of bit 35/36
				if (al & 1 << bit)
					al |= a_prop_bit;	// encroached by 1
				else
					al |= a_carry_bit;	// encroached by 0
			}
			if (r > lim)
				al |= a_carry_bit;		// encroached by 0 right of radix point
			if (subword == 5 && l & bit5)
				al |= bit8;				// flags: L35 was set for ASL
			break;

		case 's':
			al = sbox_unquote(alpha_sbox[subword][r][l]);
			break;

		case 'S':
			al = sbox_invert(gamma_sbox[subword][r], l);
			break;

		case 'u':
			/* TODO simple unary */
			break;

		case 'w':
			/* TODO stacked unary */
			break;

		case 'n':
			for (t = c32; t; t >>= 1)
				if (r & t) {
					l |= a_carry_bit;
					--t;
					break;
				}
			al = l & ~t | r & t;
			break;

		case ' ':
			break;			/* unallocated */

		default:
			oops("alpha name %c", name);
	}

	return al;
}

/* single value in sign-handling alpha5 RAM */
int alpha5_operation(int slot, int l, int r)
{
	int al = 0, range = 0;

	ifnot (slot & a5_yes)
		oops("alpha5_operation bad call");

	// Extend signs of any negative inputs.
	if (slot & a5_l_signed && l & c32)
		l |= ~63;
	if (slot & a5_r_signed && r & c32)
		r |= ~63;

	switch (slot & a5_casr) {

		case a5_cmp:						/* signed comparison */
			al = l & c63;					/* un-extend sign in result */
			if (l < r)						/* compute comparison result */
				al |= a_carry_bit;
			else if (l > r)
				al |= a_prop_bit;

			/* overflow and negative tests (used by MIN and MAX) */
			range = 5 * encode_range(r) + encode_range(l);
			break;

		case a5_rsub:						/* reverse subtract */
			l ^= r, r ^= l, l ^= r;
			nobreak;

		case a5_sub:						/* subtract */
			r = ~r;							/* method of complements */
			nobreak;

		case a5_add:
			al = l + r;

			/* overflow and negative tests */
			range = 5 * encode_range(al + 1) + encode_range(al);

			al &= c63;
			if (al == c63)					/* is carry uncertain? */
				al |= a_prop_bit;			/* yes, carries propagate */
			break;
	}

	return al | range << 8;
}

uint16_t alpha_calc(int subword, int i)
{
	int l, r, slot;

	if (use_ram)
		return alpha[subword][i];

	l = i & 63;
	r = i >> 6 & 63;
	slot = i >> 12;

	if (subword == 5 && slot >= 0x10)
		return alpha5_operation(slot, l, r);

	return alpha_operation(slot, subword, l, r);
}

/* create entire alpha RAM */
void setup_alpha(void)
{
	static int done;
	int i, sw;

	if (done++)
		oops("double setup_alpha");

	/* see "SRAM input bit assignments" table from PDF */
	for (sw=0; sw<c6; sw++)
		for (i = 0; i < 1<<16; i++)
			alpha[sw][i] = alpha_calc(sw, i);

	/* see "alpha operations dealing with signedness" table from PDF */
	for (i = 1<<16; i < 1<<17; i++)
		alpha[5][i] = alpha_calc(5, i);
}

/* simulate the alpha layer */
long alpha_sim(int oper8n, int op5, long l, long r,
	int *props, int *carries, int *t_mux, int *n_mux)
{
	int subword, q, p = 0, c = 0, lt, rt;
	long res = 0;

	oper8n <<= c12;

	for (subword = 0; subword < c6; subword++) {
		lt = c63 & l >> c6 * subword;
		rt = c63 & r >> c6 * subword;
		q = alpha_calc(subword, oper8n | rt << c6 | lt);
		res |= (long) (q & c63) << c6 * subword;
		if (q & a_prop_bit)
			p |= 1 << subword;
		if (q & a_carry_bit)
			c |= 1 << subword;
		if (subword == 4)
			oper8n = op5 << c12;			/* for last time through */
		if (subword == 5) {
			*t_mux = q >> 8 & 15;			/* temporal flag computations */
			*n_mux = q >> 12 & 3;			/* negative flag */
		}
	}

	*props = p, *carries = c;
	return res;
}

/* FOOTNOTES

A5:	a5.= in low memory is used for bitwise boolean, LSL, LSR, ROL, PIT, PAT,
	COPY, LEAD, TRAIL, and ASR. Of these, generated propagate and carries are
	ignored except for LEAD, TRAIL, and ASR. LEAD and TRAIL present a conflict,
	so these opcodes are being reimplemented as stacked unary, which won't use
	a5.=. That leaves the propagate and carry signals from a5.= available for
	exclusive use by ASR.

N:	L is negative if we're doing ASR.ss or ASR.su.
	Setting carry only affects theta.n, used only by ASR.ss and ASR.su.
	This carry is the sign bit (l & 32), copied by theta to all gamma.l
	so the sign can be extended the correct number of bits in gamma.

*/
