#include "vm.h"

// Replace the subword of width 'wi' at rightmost bit 'd' of uint 'dst' with
// the corresponding subword at rightmost bit 's' of uint/ulong 'src'.
// No operators in arguments, please: I put NO redundant parens here.
// This is designed to be maximally efficient: d, s, and w are constants.
// GCC's assembler is merely: shift, xor, and, xor.
#define w(dst, d, src, s, wi) \
	dst ^= (1 << wi) - 1 << d & (dst ^ (d > s ? src << d - s : src >> s - d));

// Same as w(), but supports ulong 'dst'. Use w() otherwise.
#define wl(dst, d, src, s, wi) \
	dst ^= (1L << wi) - 1 << d & \
	(dst ^ (d > s ? (ulong) src << d - s : (ulong) src >> s - d));

// offsets into alloc'd memory for each RAM, rather than one pointer per
enum {
	alpha0 = 0,
	alpha1 = alpha0 + bit16,
	alpha2 = alpha1 + bit16,
	alpha3 = alpha2 + bit16,
	alpha4 = alpha3 + bit16,
	alpha5 = alpha4 + bit16,		// 2 Mibit
	beta0  = alpha5 + bit16 * 2,
	beta1  = beta0  + bit16,
	beta2  = beta1  + bit16,
	beta3  = beta2  + bit16,
	beta4  = beta3  + bit16,
	beta5  = beta4  + bit16,
	gamma0 = beta5  + bit16,
	gamma1 = gamma0 + bit16,
	gamma2 = gamma1 + bit16,
	gamma3 = gamma2 + bit16,
	gamma4 = gamma3 + bit16,
	gamma5 = gamma4 + bit16,
	theta  = gamma5 + bit16,
	zeta   = theta  + bit16,		// zeta is a flag control RAM
	ram_sz = zeta   + bit16,
};

static uint16_t *Ram;

// input and output connections for all those RAMs
static uint a0i, a0o, a1i, a1o, a2i, a2o, a3i, a3o, a4i, a4o, a5i, a5o;
static uint b0i, b0o, b1i, b1o, b2i, b2o, b3i, b3o, b4i, b4o, b5i, b5o;
static uint g0i, g0o, g1i, g1o, g2i, g2o, g3i, g3o, g4i, g4o, g5i, g5o;
static uint thi, tho, zei, zeo;		// theta, flags

// left and right register outputs, 36 bits each
static ulong lro, rro;

// control decoder input (9 bits), output (max 32 bits this imp'n)
static uint cdi, cdo;

// flags input and output, with NRTZI in bits 01234
static uint fli, flo;

// glue logic inputs and outputs
// Limit one gate per integer bank, to assure no timing races.
static uint nand1a, nand1b, nand1y;
static uint nand8i = bit7, nand8o;			// 8-input NAND for active-low Z
static uint xor1a, xor1b, xor1y;			// T input flipper
static uint mux1a, mux1b, mux1s, mux1y;		// Z substitution by gamma35
static uint xor2a, xor2b, xor2y;			// Z output flipper

// wl macro with sanitized inputs
ulong wire(ulong dst, int d, ulong src, int s, int wi)
{
	wl(dst, d, src, s, wi)
	return dst;
}

void setup_alu(void)
{
	int i;

I:	Ram = calloc(ram_sz, sizeof(*Ram));
	if (!Ram)
		oops("Ram alloc");

	for (i = 0; i < bit16; i++) {
		(Ram + alpha0)[i] = alpha_calc(0, i);
		(Ram + alpha1)[i] = alpha_calc(1, i);
		(Ram + alpha2)[i] = alpha_calc(2, i);
		(Ram + alpha3)[i] = alpha_calc(3, i);
		(Ram + alpha4)[i] = alpha_calc(4, i);
		(Ram + alpha5)[i] = alpha_calc(5, i);
		(Ram +  beta0)[i] =  beta_calc(0, i);
		(Ram +  beta1)[i] =  beta_calc(1, i);
		(Ram +  beta2)[i] =  beta_calc(2, i);
		(Ram +  beta3)[i] =  beta_calc(3, i);
		(Ram +  beta4)[i] =  beta_calc(4, i);
		(Ram +  beta5)[i] =  beta_calc(5, i);
		(Ram + gamma0)[i] = gamma_calc(0, i);
		(Ram + gamma1)[i] = gamma_calc(1, i);
		(Ram + gamma2)[i] = gamma_calc(2, i);
		(Ram + gamma3)[i] = gamma_calc(3, i);
		(Ram + gamma4)[i] = gamma_calc(4, i);
		(Ram + gamma5)[i] = gamma_calc(5, i);
		(Ram + theta )[i] = theta_calc(i);
		(Ram + zeta  )[i] = flags_calc(i);
	}

	for (i = bit16; i < bit17; i++)
		(Ram + alpha5)[i] = alpha_calc(5, i);
}

// This works exclusively via global variables, because it's meant to
// operate on the /state/ of the hardware. It will grow to encompass most
// or all of the CPU.
void alu_sim(void)
{
	// control decoder RAM actuates
	cdo = DecoderRAM[cdi];

	// feed back flags from prior operation
	fli = flo;

	// NAND XOR the previous T bit to theta.
	// There's ample time for this while the register file and alpha work.
	w(nand1a, 0, fli, 2, 1)
	w(nand1b, 0, cdo, 17, 1)
	nand1y = 1 & ~(nand1a & nand1b);	// actuate NAND
	w(xor1a, 0, nand1y, 0, 1)
	w(xor1b, 0, cdo, 18, 1)
	xor1y = xor1a ^ xor1b;				// actuate XOR1
	w(thi, 12, xor1y, 0, 1)

	// operation to alpha
	w(a0i, 12, cdo, 0, 4)	w(a1i, 12, cdo, 0, 4)
	w(a2i, 12, cdo, 0, 4)	w(a3i, 12, cdo, 0, 4)
	w(a4i, 12, cdo, 0, 4)	w(a5i, 12, cdo, 0, 2)	w(a5i, 14, cdo, 4, 3)

	// operation to beta
	w(b0i, 12, cdo, 7, 4)	w(b1i, 12, cdo, 7, 4)	w(b2i, 12, cdo, 7, 4)
	w(b3i, 12, cdo, 7, 4)	w(b4i, 12, cdo, 7, 4)	w(b5i, 12, cdo, 7, 4)

	// operation to gamma
	w(g0i, 13, cdo, 11, 3)	w(g1i, 13, cdo, 11, 3)	w(g2i, 13, cdo, 11, 3)
	w(g3i, 13, cdo, 11, 3)	w(g4i, 13, cdo, 11, 3)	w(g5i, 13, cdo, 11, 3)

	// operation to theta and flags
	w(thi, 13, cdo, 14, 3)
	w(zei, 11, cdo, 19, 4)

	// left operand to alpha
	w(a0i, 0, lro,  0, 6)
	w(a1i, 0, lro,  6, 6)
	w(a2i, 0, lro, 12, 6)
	w(a3i, 0, lro, 18, 6)
	w(a4i, 0, lro, 24, 6)
	w(a5i, 0, lro, 30, 6)

	// right operand to alpha, beta, gamma
	w(a0i, 6, rro,  0, 6)	w(b0i, 6, rro,  0, 6)	w(g0i, 6, rro,  0, 6)
	w(a1i, 6, rro,  6, 6)	w(b1i, 6, rro,  6, 6)	w(g1i, 6, rro,  6, 6)
	w(a2i, 6, rro, 12, 6)	w(b2i, 6, rro, 12, 6)	w(g2i, 6, rro, 12, 6)
	w(a3i, 6, rro, 18, 6)	w(b3i, 6, rro, 18, 6)	w(g3i, 6, rro, 18, 6)
	w(a4i, 6, rro, 24, 6)	w(b4i, 6, rro, 24, 6)	w(g4i, 6, rro, 24, 6)
	w(a5i, 6, rro, 30, 6)	w(b5i, 6, rro, 30, 6)	w(g5i, 6, rro, 30, 6)

	// alpha RAMs actuate
	a0o = (Ram + alpha0)[a0i];
	a1o = (Ram + alpha1)[a1i];
	a2o = (Ram + alpha2)[a2i];
	a3o = (Ram + alpha3)[a3i];
	a4o = (Ram + alpha4)[a4i];
	a5o = (Ram + alpha5)[a5i];

	// alpha --> transpose --> beta
T:	w(b0i, 0, a0o, 0, 1)	w(b1i, 0, a0o, 1, 1)	w(b2i, 0, a0o, 2, 1)
	w(b0i, 1, a1o, 0, 1)	w(b1i, 1, a1o, 1, 1)	w(b2i, 1, a1o, 2, 1)
	w(b0i, 2, a2o, 0, 1)	w(b1i, 2, a2o, 1, 1)	w(b2i, 2, a2o, 2, 1)
	w(b0i, 3, a3o, 0, 1)	w(b1i, 3, a3o, 1, 1)	w(b2i, 3, a3o, 2, 1)
	w(b0i, 4, a4o, 0, 1)	w(b1i, 4, a4o, 1, 1)	w(b2i, 4, a4o, 2, 1)
	w(b0i, 5, a5o, 0, 1)	w(b1i, 5, a5o, 1, 1)	w(b2i, 5, a5o, 2, 1)

	w(b3i, 0, a0o, 3, 1)	w(b4i, 0, a0o, 4, 1)	w(b5i, 0, a0o, 5, 1)
	w(b3i, 1, a1o, 3, 1)	w(b4i, 1, a1o, 4, 1)	w(b5i, 1, a1o, 5, 1)
	w(b3i, 2, a2o, 3, 1)	w(b4i, 2, a2o, 4, 1)	w(b5i, 2, a2o, 5, 1)
	w(b3i, 3, a3o, 3, 1)	w(b4i, 3, a3o, 4, 1)	w(b5i, 3, a3o, 5, 1)
	w(b3i, 4, a4o, 3, 1)	w(b4i, 4, a4o, 4, 1)	w(b5i, 4, a4o, 5, 1)
	w(b3i, 5, a5o, 3, 1)	w(b4i, 5, a5o, 4, 1)	w(b5i, 5, a5o, 5, 1)

	// beta RAMs actuate
	b0o = (Ram + beta0)[b0i];
	b1o = (Ram + beta1)[b1i];
	b2o = (Ram + beta2)[b2i];
	b3o = (Ram + beta3)[b3i];
	b4o = (Ram + beta4)[b4i];
	b5o = (Ram + beta5)[b5i];

	// carry and propagate from alpha to theta
	w(thi, 0, a0o, 6, 1)	w(thi,  6, a0o, 7, 1)
	w(thi, 1, a1o, 6, 1)	w(thi,  7, a1o, 7, 1)
	w(thi, 2, a2o, 6, 1)	w(thi,  8, a2o, 7, 1)
	w(thi, 3, a3o, 6, 1)	w(thi,  9, a3o, 7, 1)
	w(thi, 4, a4o, 6, 1)	w(thi, 10, a4o, 7, 1)
	w(thi, 5, a5o, 6, 1)	w(thi, 11, a5o, 7, 1)

	// theta RAM actuates
	tho = (Ram + theta)[thi];

	// beta --> transpose --> gamma
	w(g0i, 0, b0o, 0, 1)	w(g1i, 0, b0o, 1, 1)	w(g2i, 0, b0o, 2, 1)
	w(g0i, 1, b1o, 0, 1)	w(g1i, 1, b1o, 1, 1)	w(g2i, 1, b1o, 2, 1)
	w(g0i, 2, b2o, 0, 1)	w(g1i, 2, b2o, 1, 1)	w(g2i, 2, b2o, 2, 1)
	w(g0i, 3, b3o, 0, 1)	w(g1i, 3, b3o, 1, 1)	w(g2i, 3, b3o, 2, 1)
	w(g0i, 4, b4o, 0, 1)	w(g1i, 4, b4o, 1, 1)	w(g2i, 4, b4o, 2, 1)
	w(g0i, 5, b5o, 0, 1)	w(g1i, 5, b5o, 1, 1)	w(g2i, 5, b5o, 2, 1)

	w(g3i, 0, b0o, 3, 1)	w(g4i, 0, b0o, 4, 1)	w(g5i, 0, b0o, 5, 1)
	w(g3i, 1, b1o, 3, 1)	w(g4i, 1, b1o, 4, 1)	w(g5i, 1, b1o, 5, 1)
	w(g3i, 2, b2o, 3, 1)	w(g4i, 2, b2o, 4, 1)	w(g5i, 2, b2o, 5, 1)
	w(g3i, 3, b3o, 3, 1)	w(g4i, 3, b3o, 4, 1)	w(g5i, 3, b3o, 5, 1)
	w(g3i, 4, b4o, 3, 1)	w(g4i, 4, b4o, 4, 1)	w(g5i, 4, b4o, 5, 1)
	w(g3i, 5, b5o, 3, 1)	w(g4i, 5, b5o, 4, 1)	w(g5i, 5, b5o, 5, 1)

	// carry decisions from theta to gamma
	w(g0i, 12, tho, 0, 1)	w(g1i, 12, tho, 1, 1)	w(g2i, 12, tho, 2, 1)
	w(g3i, 12, tho, 3, 1)	w(g4i, 12, tho, 4, 1)	w(g5i, 12, tho, 5, 1)

	// flag parameters to zeta (already got control decoder flag scheme)
	w(zei,  0, fli, 0, 4)
	w(zei,  4, a5o, 8, 5)
	w(zei,  9, tho, 6, 1)
	w(zei, 10, tho, 7, 1)

	// gamma RAMs actuate
	g0o = (Ram + gamma0)[g0i];
	g1o = (Ram + gamma1)[g1i];
	g2o = (Ram + gamma2)[g2i];
	g3o = (Ram + gamma3)[g3i];
	g4o = (Ram + gamma4)[g4i];
	g5o = (Ram + gamma5)[g5i];

	// zeta RAM actuates
	zeo = (Ram + zeta)[zei];

	// 8-input NAND has one input tied to Vdd. Connect remainder.
	w(nand8i, 0, g0o, 6, 1)
	w(nand8i, 1, g1o, 6, 1)
	w(nand8i, 2, g2o, 6, 1)
	w(nand8i, 3, g3o, 6, 1)
	w(nand8i, 4, g4o, 6, 1)
	w(nand8i, 5, g5o, 6, 1)
	w(nand8i, 6, zeo, 3, 1)
	nand8o = nand8i == 255 ? 0 : 1;		// NAND8 actuates; active-low Z flag

	// There's an XOR after NAND8 to override zero flag.
	w(xor2a, 0, nand8o, 0, 1)
	w(xor2b, 0, zeo, 6, 1)
	xor2y = xor2a ^ xor2b;				// actuate XOR2

	// mux chooses between flag RAM's N flag and copy of gamma35
	w(mux1a, 0, zeo, 0, 1)
	w(mux1b, 0, g5o, 7, 1)
	w(mux1s, 0, zeo, 5, 1)
	mux1y = mux1s ? mux1b : mux1a;		// actuate MUX1

	// output flags
	w(flo, 0, mux1y, 0, 1)
	w(flo, 1, zeo, 1, 2)
	w(flo, 3, xor2y, 0, 1)
	w(flo, 4, zeo, 4, 1)
}

/* show ALU instrument panel */
static void glow(long y)
{
	uint a_op = 0, b_op = 0, g_op = 0, t_op = 0;
	uint t_cooked_in = 0;

	w(a_op, 0, cdo,  0, 4)		// XXX Keep in sync with actual ALU.
	w(b_op, 0, cdo,  7, 4)
	w(g_op, 0, cdo, 11, 3)
	w(t_op, 0, cdo, 14, 3)
	w(t_cooked_in, 0, xor1y, 0, 1)

	print36("\n        L " RED, lro, BAW "   ");
	printf("  " CYA "%c" BAW " -> " CYA "%c" BAW "\n",
		flo & bit1 ? 'T' : 't', xor1y & bit0 ? 'T' : 't');
	print36("        R " RED, rro, BAW "\n");

	printf(" %c  alpha ", alpha_slots[a_op]);
	print6(GRE,  a5o, " ");	print6(NULL, a4o, " ");	print6(NULL, a3o, " ");
	print6(NULL, a2o, " ");	print6(NULL, a1o, " ");	print6(NULL, a0o, BAW);
	print6("   p " CYA, 63 & thi >> 6, BAW "\n");

	printf(" %c   beta ", beta_slots[b_op]);
	print6(GRE,  b5o, " ");	print6(NULL, b4o, " ");	print6(NULL, b3o, " ");
	print6(NULL, b2o, " ");	print6(NULL, b1o, " ");	print6(NULL, b0o, BAW);
	print6("   c " CYA, 63 & thi, BAW "\n");

	printf(" %.2s gamma ", gamma_slots + 2 * g_op);
	print6(GRE,  g5o, " ");	print6(NULL, g4o, " ");	print6(NULL, g3o, " ");
	print6(NULL, g2o, " ");	print6(NULL, g1o, " ");	print6(NULL, g0o, BAW);
	print6("   d " CYA, 63 & tho, BAW);
	printf(" %c\n", theta_slots[2 * t_op + t_cooked_in]);

	printf("    flags " CYA "%.1s %.1s %.1s %.1s %.1s" BAW "\n",
		"Ii" + !(flo & bit4),
		"Nn" + !(flo & bit0), "Rr" + !(flo & bit1),
		"Tt" + !(flo & bit2), "zZ" + !(flo & bit3));	// Z is active low.

	printf(" unsigned " CYA "%li" BAW "\n", y);
	printf("   signed " CYA "%li" BAW "\n", y & pole ? y | comp36 : y);

	if (getenv("ALU_HUH")) {
#define what(huh) printf(#huh " %o\n", huh);
		puts("");
		what(fli)
		what(flo)
		what(nand1a)
		what(nand1b)
		what(nand1y)
		what(nand8i)
		what(nand8o)
		what(xor1a)
		what(xor1b)
		what(xor1y)
		what(mux1a)
		what(mux1b)
		what(mux1s)
		what(mux1y)
		what(a5o)
		what(zei)
		what(zeo)
		what(thi)
		what(tho)
#undef what
	}
}

long alu(int opcode, long l, long r, int *flags)
{
	long y = 0;

	cdi = opcode;
	lro = l;
	rro = r;
	flo = flags ? *flags : 0;			// immediately fed to fli
	alu_sim();

	wl(y,  0, g0o, 0, 6)
	wl(y,  6, g1o, 0, 6)
	wl(y, 12, g2o, 0, 6)
	wl(y, 18, g3o, 0, 6)
	wl(y, 24, g4o, 0, 6)
	wl(y, 30, g5o, 0, 6)

	if (show)
		glow(y);

	if (flags) *flags = flo;
	return y;
}

/*	FOOTNOTES

I:	Unused labels do not generate code with -O0 or -O3,
	so they are harmless for annotation of statements.

T:	Our "sliding word" transpose() using a long generates fewer instructions,
	but does not include packing and unpacking the long. I suspect this
	construct is faster, based on looking at -O0 and -O3 assembly output
	from both.

*/
