/* beta SRAM definition

INPUT BITS

	0-5		transposed input from a0..a5
	6-11	right input
	12-15	operation

OUTPUT BITS

	0-5		transposed output to g0..g5

*/

#include "vm.h"

uint8_t beta[6][1 << 16];

int beta_operation(int slot, int subword, int l, int r)
{
	int b = 0, bit, mask;
	int name = beta_slots[slot];
	int sign_ext = 0;

	if (l<0 || r<0 || l>c63 || r > c63 || subword < 0 || subword > c5)
		oops("beta_operation");

	/* see "beta operations" table from PDF */
	switch (name) {
		case '=':
			b = l;
			break;

		case 'L':
			for (mask = bit = 0; bit < 6; ++bit)
				if (r <= 6 * (5 - bit) + 5 - subword)
					mask |= 1 << bit;

			l &= mask;
			nobreak;

		case 'l':								/* rotate word left */
l:			r %= 36;							/* word rotation, 0-35 */
			r = (36 + r + subword) / 6 % 6;		/* subword's rotation, 0-5 */
			b = rotate_tribble_left(l, r);		/* rotated subword */
			break;

		case '}':
			sign_ext = l & 32 ? 63 : 0;
			nobreak;

		case 'R':
			if (!r)								/* no shift? */
				goto l;
			for (mask = bit = 0; bit < 6; ++bit)
				if (r < 36 && r >= 36 - 6 * bit - subword)
					mask |= 1 << bit;
			l = l & mask | sign_ext & ~mask;
			goto l;

		case 's':
			b = sbox_unquote(beta_sbox[subword][r][l]);
			break;

		case 'S':
			b = sbox_invert(beta_sbox[subword][r], l);
			break;

		case 'w':
			/* TODO stacked unary */
			break;

		case '!':
			b = c63 & ~l;
			break;

		case '^':
			b = c63 & (l ^ r);
			break;

		case '*':
			b = c63 & l << 1 | l >> c5;
			break;

		case 'p':
			b = permute(r, l);
			break;

		case 'z':
			/* XXX only the first 6 swizzles are defined */
			if (r < c6)
				b = l & 1 << r ? c63 : 0;
			break;

		case ' ':			/* unallocated slots */
			b = 0;
			break;

		default:
			oops("beta name '%c' slot %i", name, slot);
	}

	return b;
}

uint16_t beta_calc(int subword, int i)
{
	int l, r, slot;

	if (use_ram)
		return beta[subword][i];

	l = i & 63;
	r = i >> 6 & 63;
	slot = i >> 12;

	return beta_operation(slot, subword, l, r);
}

/* create entire beta RAM */
void setup_beta(void)
{
	static int done;
	int i, sw;

	if (done++)
		oops("double setup_beta");

	/* see "SRAM input bit assignments" table from PDF */
	for (sw=0; sw<c6; sw++)
		for (i = 0; i < 1<<16; i++)
			beta[sw][i] = beta_calc(sw, i);
}

/* transpose 36-bit word (uses almost full 64-bit width) */
/* ZYXWVU TSRQPO NMLKJI HGFEDC BA9876 543210 --> */
/* ZTNHB5 YSMGA4 XRLF93 WQKE82 VPJD71 UOIC60 */
long transpose(long w)
{
	ulong x, y = 0;

	/* bit 0 in aligned to bit 0 out - constants are octal */
	x = w & 010101010101;
	y |= 077 & (x >> 25 | x >> 20 | x >> 15 | x >> 10 | x >> 5 | x);

	/* bit 1 in aligned to bit 6 out */
	x = (w & 020202020202) << 5;
	y |= 07700 & (x >> 25 | x >> 20 | x >> 15 | x >> 10 | x >> 5 | x);

	/* bit 2 in aligned to bit 12 out */
	x = (w & 040404040404) << 10;
	y |= 0770000 & (x >> 25 | x >> 20 | x >> 15 | x >> 10 | x >> 5 | x);

	/* bit 3 in aligned to bit 18 out */
	x = (w & 0101010101010) << 15;
	y |= 077000000 & (x >> 25 | x >> 20 | x >> 15 | x >> 10 | x >> 5 | x);

	/* bit 4 in aligned to bit 24 out */
	x = (w & 0202020202020) << 20;
	y |= 07700000000 & (x >> 25 | x >> 20 | x >> 15 | x >> 10 | x >> 5 | x);

	/* bit 5 in aligned to bit 30 out */
	x = (w & 0404040404040) << 25;
	y |= 0770000000000 & (x >> 25 | x >> 20 | x >> 15 | x >> 10 | x >> 5 | x);

	return y;
}

/* Example of wire() function. Entire ALU runs 3.5x slower using this. */
ulong transpose_slow(ulong w)
{
	return slow_wire("ztnhb5 ysmga4 xrlf93 wqke82 vpjd71 uoic60", w);
}

/* simulate the beta layer */
long beta_sim(int operation, long l, long r)
{
	int subword, q, lt, rt;
	long res = 0;

	l = transpose(l);
	operation <<= c12;

	for (subword=0; subword<c6; subword++) {
		lt = c63 & l >> c6 * subword;
		rt = c63 & r >> c6 * subword;
		q = beta_calc(subword, operation | rt << c6 | lt);
		res |= (long) (q & c63) << c6 * subword;
	}

	return transpose(res);
}
