/* Control decoder data (at least)

OUTPUT BITS

	0-3		a0..a4 operation
	0-1		a5 operation bits 0-1
	4-6		a5 operation bits 2-4
	7-10	beta operation
	11-13	gamma operation
	14-16	theta operation
T:	17		pass T into ALU (set), pass 1 instead of T (clear)
T:	18		input T as-is (set), invert T before theta (clear)
	19-22	flag scheme

	XXX		to be continued

FLAG SEMANTICS AS WE SEE THEM

	The first matching rule in this list prevails.

	1.	LOAD and STORE do not change any flags.
	2.	BOUND only changes I flag.
	3.	N and Z are changed by all remaining instructions,
		but N will not always match gamma_35.
	4.	R is set whenever T is set, and stays set until purposely cleared.
	5.	T is altered by additive, arithmetic shift, and min/max families only.

 */

#include "vm.h"

char *alpha_slots = "=+-m&|xX*nsSuweE";
char * beta_slots = "=lLR}sSw!^*pz   ";
char *gamma_slots = "??++kkwwsSu^lLp ";		/* must divide by 2 */
char *theta_slots = "++<>{}01n       ";		/* also divide by 2 */

uint *DecoderRAM;

struct decoder Decoder[oc_NO_MORE] = {
/*  OPCODE       NAME       A    B    G    T     A5  t_v  f_sch */
	oc_A_uuu,   "a.uuu"  , '+', '=', '+', '+', a5uu, t_0, fs_add_u,
	oc_A_uus,   "a.uus"  , '+', '=', '+', '+', a5uu, t_0, fs_add_s,
	oc_A_usu,   "a.usu"  , '+', '=', '+', '+', a5us, t_0, fs_add_u,
	oc_A_uss,   "a.uss"  , '+', '=', '+', '+', a5us, t_0, fs_add_s,
	oc_A_suu,   "a.suu"  , '+', '=', '+', '+', a5su, t_0, fs_add_u,
	oc_A_sus,   "a.sus"  , '+', '=', '+', '+', a5su, t_0, fs_add_s,
	oc_A_ssu,   "a.ssu"  , '+', '=', '+', '+', a5ss, t_0, fs_add_u,
	oc_A_sss,   "a.sss"  , '+', '=', '+', '+', a5ss, t_0, fs_add_s,

	oc_S_uuu,   "s.uuu"  , '-', '=', '+', '+', a5uu, t_1, fs_add_u,
	oc_S_uus,   "s.uus"  , '-', '=', '+', '+', a5uu, t_1, fs_add_s,
	oc_S_usu,   "s.usu"  , '-', '=', '+', '+', a5us, t_1, fs_add_u,
	oc_S_uss,   "s.uss"  , '-', '=', '+', '+', a5us, t_1, fs_add_s,
	oc_S_suu,   "s.suu"  , '-', '=', '+', '+', a5su, t_1, fs_add_u,
	oc_S_sus,   "s.sus"  , '-', '=', '+', '+', a5su, t_1, fs_add_s,
	oc_S_ssu,   "s.ssu"  , '-', '=', '+', '+', a5ss, t_1, fs_add_u,
	oc_S_sss,   "s.sss"  , '-', '=', '+', '+', a5ss, t_1, fs_add_s,

	oc_RS_uuu,  "rs.uuu" , 'm', '=', '+', '+', a5uu, t_1, fs_add_u,
	oc_RS_uus,  "rs.uus" , 'm', '=', '+', '+', a5uu, t_1, fs_add_s,
	oc_RS_usu,  "rs.usu" , 'm', '=', '+', '+', a5us, t_1, fs_add_u,
	oc_RS_uss,  "rs.uss" , 'm', '=', '+', '+', a5us, t_1, fs_add_s,
	oc_RS_suu,  "rs.suu" , 'm', '=', '+', '+', a5su, t_1, fs_add_u,
	oc_RS_sus,  "rs.sus" , 'm', '=', '+', '+', a5su, t_1, fs_add_s,
	oc_RS_ssu,  "rs.ssu" , 'm', '=', '+', '+', a5ss, t_1, fs_add_u,
	oc_RS_sss,  "rs.sss" , 'm', '=', '+', '+', a5ss, t_1, fs_add_s,

	oc_AC_uuu,  "ac.uuu" , '+', '=', '+', '+', a5uu, t_n, fs_add_u,
	oc_AC_uus,  "ac.uus" , '+', '=', '+', '+', a5uu, t_n, fs_add_s,
	oc_AC_usu,  "ac.usu" , '+', '=', '+', '+', a5us, t_n, fs_add_u,
	oc_AC_uss,  "ac.uss" , '+', '=', '+', '+', a5us, t_n, fs_add_s,
	oc_AC_suu,  "ac.suu" , '+', '=', '+', '+', a5su, t_n, fs_add_u,
	oc_AC_sus,  "ac.sus" , '+', '=', '+', '+', a5su, t_n, fs_add_s,
	oc_AC_ssu,  "ac.ssu" , '+', '=', '+', '+', a5ss, t_n, fs_add_u,
	oc_AC_sss,  "ac.sss" , '+', '=', '+', '+', a5ss, t_n, fs_add_s,

	oc_SC_uuu,  "sc.uuu" , '-', '=', '+', '+', a5uu, t_i, fs_add_u,
	oc_SC_uus,  "sc.uus" , '-', '=', '+', '+', a5uu, t_i, fs_add_s,
	oc_SC_usu,  "sc.usu" , '-', '=', '+', '+', a5us, t_i, fs_add_u,
	oc_SC_uss,  "sc.uss" , '-', '=', '+', '+', a5us, t_i, fs_add_s,
	oc_SC_suu,  "sc.suu" , '-', '=', '+', '+', a5su, t_i, fs_add_u,
	oc_SC_sus,  "sc.sus" , '-', '=', '+', '+', a5su, t_i, fs_add_s,
	oc_SC_ssu,  "sc.ssu" , '-', '=', '+', '+', a5ss, t_i, fs_add_u,
	oc_SC_sss,  "sc.sss" , '-', '=', '+', '+', a5ss, t_i, fs_add_s,

	oc_RSC_uuu, "rsc.uuu", 'm', '=', '+', '+', a5uu, t_i, fs_add_u,
	oc_RSC_uus, "rsc.uus", 'm', '=', '+', '+', a5uu, t_i, fs_add_s,
	oc_RSC_usu, "rsc.usu", 'm', '=', '+', '+', a5us, t_i, fs_add_u,
	oc_RSC_uss, "rsc.uss", 'm', '=', '+', '+', a5us, t_i, fs_add_s,
	oc_RSC_suu, "rsc.suu", 'm', '=', '+', '+', a5su, t_i, fs_add_u,
	oc_RSC_sus, "rsc.sus", 'm', '=', '+', '+', a5su, t_i, fs_add_s,
	oc_RSC_ssu, "rsc.ssu", 'm', '=', '+', '+', a5ss, t_i, fs_add_u,
	oc_RSC_sss, "rsc.sss", 'm', '=', '+', '+', a5ss, t_i, fs_add_s,

	oc_IGF,     "igf"    , '=', '=', 'k', '1', 0x00, t_1, fs_logic,
	oc_IGT,     "igt"    , '=', '=', '?', '1', 0x00, t_1, fs_logic,
	oc_XL,      "xl"     , '=', '=', '?', '0', 0x00, t_0, fs_logic,
	oc_XR,      "xr"     , '=', '=', '?', '1', 0x00, t_1, fs_logic,
	oc_NL,      "nl"     , '=', '!', '+', '0', 0x00, t_0, fs_logic,
	oc_NR,      "nr"     , '=', '=', '^', '1', 0x00, t_1, fs_logic,
	oc_AND,     "and"    , '&', '=', '+', '0', 0x00, t_0, fs_logic,
	oc_NAND,    "nand"   , '&', '!', '+', '0', 0x00, t_0, fs_logic,
	oc_OR,      "or"     , '|', '=', '+', '0', 0x00, t_0, fs_logic,
	oc_NOR,     "nor"    , '|', '!', '+', '0', 0x00, t_0, fs_logic,
	oc_XOR,     "xor"    , '=', '=', '^', '1', 0x00, t_1, fs_logic,
	oc_XNOR,    "xnor"   , '=', '!', '^', '1', 0x00, t_1, fs_logic,
	oc_LANR,    "lanr"   , '|', '=', '^', '1', 0x00, t_1, fs_logic,
	oc_RANL,    "ranl"   , '&', '=', '^', '1', 0x00, t_1, fs_logic,
	oc_LONR,    "lonr"   , '&', '!', '^', '1', 0x00, t_1, fs_logic,
	oc_RONL,    "ronl"   , '|', '!', '^', '1', 0x00, t_1, fs_logic,

	oc_MAX_uuu, "max.uuu", '=', '=', '?', '<', a5uu, t_0, fs_add_u,
	oc_MAX_uus, "max.uus", '=', '=', '?', '<', a5uu, t_0, fs_add_s,
	oc_MAX_usu, "max.usu", '=', '=', '?', '<', a5us, t_0, fs_add_u,
	oc_MAX_uss, "max.uss", '=', '=', '?', '<', a5us, t_0, fs_add_s,
	oc_MAX_suu, "max.suu", '=', '=', '?', '<', a5su, t_0, fs_add_u,
	oc_MAX_sus, "max.sus", '=', '=', '?', '<', a5su, t_0, fs_add_s,
	oc_MAX_ssu, "max.ssu", '=', '=', '?', '<', a5ss, t_0, fs_add_u,
	oc_MAX_sss, "max.sss", '=', '=', '?', '<', a5ss, t_0, fs_add_s,

	oc_MIN_uuu, "min.uuu", '=', '=', '?', '>', a5uu, t_1, fs_add_u,
	oc_MIN_uus, "min.uus", '=', '=', '?', '>', a5uu, t_1, fs_add_s,
	oc_MIN_usu, "min.usu", '=', '=', '?', '>', a5us, t_1, fs_add_u,
	oc_MIN_uss, "min.uss", '=', '=', '?', '>', a5us, t_1, fs_add_s,
	oc_MIN_suu, "min.suu", '=', '=', '?', '>', a5su, t_1, fs_add_u,
	oc_MIN_sus, "min.sus", '=', '=', '?', '>', a5su, t_1, fs_add_s,
	oc_MIN_ssu, "min.ssu", '=', '=', '?', '>', a5ss, t_1, fs_add_u,
	oc_MIN_sss, "min.sss", '=', '=', '?', '>', a5ss, t_1, fs_add_s,

	oc_BOUND_u, "bound.u", '-', '=', '+', '+', a5uu, t_1, fs_bound,
	oc_BOUND_s, "bound.s", '-', '=', '+', '+', a5us, t_1, fs_bound,
	oc_CMP_uu,  "cmp.uu" , '-', '=', '+', '+', a5uu, t_1, fs_cmp,
	oc_CMP_us,  "cmp.us" , '-', '=', '+', '+', a5us, t_1, fs_cmp,
	oc_CMP_su,  "cmp.su" , '-', '=', '+', '+', a5su, t_1, fs_cmp,
	oc_CMP_ss,  "cmp.ss" , '-', '=', '+', '+', a5ss, t_1, fs_cmp,

	oc_ASL_uu,  "asl.uu" , 'E', 'L', 'l', '0', 0x00, t_0, fs_asl_uu,
	oc_ASL_us,  "asl.us" , 'e', 'L', 'l', '0', 0x00, t_0, fs_asl_us,
	oc_ASL_su,  "asl.su" , 'E', 'L', 'l', '0', 0x00, t_0, fs_asl_su,
	oc_ASL_ss,  "asl.ss" , 'e', 'L', 'l', '0', 0x00, t_0, fs_asl_ss,

	oc_ASR_uu,  "asr.uu" , '=', 'R', 'l', '0', 0x00, t_0, fs_asr_uu,
	oc_ASR_us,  "asr.us" , '=', 'R', 'l', '0', 0x00, t_0, fs_asr_us,
	oc_ASR_su,  "asr.su" , '=', 'R', 'l', 'n', 0x00, t_0, fs_asr_su,
	oc_ASR_ss,  "asr.ss" , '=', 'R', 'l', 'n', 0x00, t_0, fs_asr_ss,
	
	oc_LSL,     "lsl"    , '=', 'L', 'l', '0', 0x00, t_0, fs_logic,
	oc_LSR,     "lsr"    , '=', 'R', 'l', '0', 0x00, t_0, fs_logic,
	oc_ROL,     "rol"    , '=', 'l', 'l', '0', 0x00, t_0, fs_logic,

	oc_ML,      "ml"     , 'x', '=', '+', '0', 0x00, t_0, fs_asr_uu,
	oc_MH,      "mh"     , 'X', '*', '+', '0', 0x00, t_0, fs_asr_us,
	oc_MHNS,    "mhns"   , 'X', '=', '+', '0', 0x00, t_0, fs_asr_uu,
	oc_DSL,     "dsl"    , '*', '*', '+', '0', 0x00, t_0, fs_asr_uu,
	oc_COPY,    "copy"   , '=', 'z', '+', '0', 0x00, t_0, fs_asr_uu,

	oc_NUDGE,   "nudge"  , 'n', '=', '?', '}', 0x00, t_1, fs_logic,
	oc_CRF,     "crf"    , '=', '=', 'k', '1', 0x00, t_1, fs_clear_r,

	oc_PIT,     "pit"    , '=', '=', 'p', '0', 0x00, t_0, fs_logic,
	oc_PAT,     "pat"    , '=', 'p', '+', '0', 0x00, t_0, fs_logic,
	oc_PAIT,    "pait"   , '=', 'p', 'p', '0', 0x00, t_0, fs_logic,

	oc_MIX,     "mix"    , 's', 's', 's', '0', 0x00, t_0, fs_logic,
	oc_XIM,     "xim"    , 'S', 'S', 'S', '1', 0x00, t_1, fs_logic,

//	oc_NO_MORE, "no.more", ' ', ' ', ' ', ' ', 0x00, t_0, fs_logic,
};

void setup_decoder(void)
{
	static int done;
	int i;
	char *f;
	uint rom;

	if (done++)
		oops("double scd");

	ifnot (DecoderRAM = calloc(512, sizeof(*DecoderRAM)))
		oops("alloc decoder");		// XXX hard const 512 problematic

	for (i=0; i<oc_NO_MORE; i++) {
		rom = 0;
		if (Decoder[i].opcode != i)
			oops("slot %i had opcode %i", i, Decoder[i].opcode);

		ifnot (f = strchr(alpha_slots, Decoder[i].alpha))
			oops("slot %i had alpha %c", i, Decoder[i].alpha);
		Decoder[i].alpha = f - alpha_slots;
		rom = wire(rom, 0, f - alpha_slots, 0, 4);

		if (!Decoder[i].alpha5)
			Decoder[i].alpha5 = Decoder[i].alpha;
		rom = wire(rom, 4, Decoder[i].alpha5, 2, 3);

		ifnot (f = strchr(beta_slots, Decoder[i].beta))
			oops("slot %i had beta %c", i, Decoder[i].beta);
		Decoder[i].beta = f - beta_slots;
		rom = wire(rom, 7, f - beta_slots, 0, 4);

		ifnot (f = strchr(gamma_slots, Decoder[i].gamma))
			oops("slot %i had gamma %c", i, Decoder[i].gamma);
		Decoder[i].gamma = (f - gamma_slots) / 2;
		rom = wire(rom, 11, (f - gamma_slots) / 2, 0, 3);

		ifnot (f = strchr(theta_slots, Decoder[i].theta))
			oops("slot %i had theta %c", i, Decoder[i].theta);
		Decoder[i].theta = (f - theta_slots) / 2;
		rom = wire(rom, 14, (f - theta_slots) / 2, 0, 3);

		rom = wire(rom, 17, Decoder[i].t_vision, 0, 2);

		rom = wire(rom, 19, Decoder[i].flag_scheme, 0, 4);

		DecoderRAM[i] = rom;
	}
}

/* FOOTNOTES

T:	These two bits are t_enable and t_not_inverted in the logic:
	carry to ALU = (t_flag NAND t_enable) XOR t_not_inverted
	Hard logic to follow, but uses the simplest hardware.

*/
