/* flags SRAM definition

INPUT BITS

	0		previous N
	1		previous R
	2		previous T
	3		previous Z (active low)
	4-8		alpha5 encoded range
	9		theta carry summary
	10		theta propagate summary
	11-14	flag scheme from control decoder
	15		- unallocated -

OUTPUT BITS
XXX This order is functionally acceptable, but not aesthetically ideal.
XXX Z no longer needs to be active low.

	0		N'egative (per additive functions)
	1		R'ange
	2		T'emporal range
	3		Z'ero allowed (active high)
	4		I'nterrupt
	5		if set, substitute gamma_35 for N flag
	6		if set, invert the final zero determination
			(zero remains active low)

*/

#include "vm.h"

// XXX There may be "safer" ways to write this routine, maybe table driven,
//     to avoid some types of oversights like not preserving Z.
static int flags_operation(
	int prev_n,
	int prev_r,
	int prev_t,
	int prev_z,
	int encoded_range,
	int c_summ,
	int p_summ,
	int scheme
) {
	int f = 0, t_changes = 1;
	int range = c_summ ? encoded_range / 5 : encoded_range % 5;

	switch (scheme) {

		case fs_logic:				// keep T; N & Z come from gamma
			f = 0x28 | prev_t << 2;
			t_changes = 0;
			break;

		case fs_cmp:				// Z from gamma, N from subtraction, keep T
			if (range < 2)
				f |= bit0;			// output is negative
			if (range == 2)
				f |= bit3;			// possible for a zero result
			f |= prev_t << 2;		// keep T
			t_changes = 0;
			break;

		case fs_add_s:
			if (range != 1 && range != 2)
				f = bit2;			// does not fit in signed result
			if (range < 2)
				f |= bit0;			// output is negative
			if (range == 2)
				f |= bit3;			// possible for a zero result
			break;

		case fs_add_u:
			if (range != 2 && range != 3)
				f = bit2;			// does not fit in unsigned result
			if (range < 2)
				f |= bit0;			// output is negative
			if (range == 2)
				f |= bit3;			// possible for a zero result
			break;

		case fs_bound:				// can only change I'nterrupt flag
			f = prev_t << 2 | prev_n;
			t_changes = 0;
Z:			if (!prev_z)
				f |= bit6;
			if (range > 1)
				f |= bit4;			// array subscript out of bounds
			break;

		case fs_clear_r:			// can only change Range flag
			f = prev_t << 2 | prev_n;
			if (!prev_z)
				f |= bit6;
			t_changes = 0;
			prev_r = 0;
			break;

		case fs_asr_uu:
			f = bit3;				// Z follows gamma, N=0, T=0
			break;

		case fs_asr_us:
			f = bit3;				// Z follows gamma, N=0
			if ((encoded_range & 3) == 3)
				f |= bit2;			// T if negative input, no shift, uns. out
			break;

		case fs_asr_su:
			f = bit3;				// Z follows gamma
			if (encoded_range & 1)
				f |= bit0 | bit2;	// N and T follows L35
			break;

		case fs_asr_ss:
			f = bit3;				// Z follows gamma
			if (encoded_range & 1)
				f |= bit0;			// N follows L35
			break;

		case fs_asl_uu:				// N=0
			if (p_summ)				// 1 encroaches bit 36?
				f |= bit2;			// set T, force Z zero
			else
				f |= bit3;			// no T, gamma may set Z
			break;

		case fs_asl_us:				// N=0
			// 1 encroaches bit 35 or input is negative?
			if (p_summ || encoded_range & 1)
				f |= bit2;			// set T, force Z zero
			else
				f |= bit3;			// no T, gamma may set Z
			break;

		case fs_asl_su:
			// 1 encroaches bit 36 or input is negative?
			if (p_summ || encoded_range & 1)
				f |= bit2;			// set T, force Z zero
			else
				f |= bit3;			// no T, gamma may set Z
			if (encoded_range & 1)
				f |= bit0;			// N follows L35
			break;

		case fs_asl_ss:
			// 1 encroaches bit 35 with non-negative input, or
			// 1 encroaches bit 35 with negative input?
			if (p_summ && !(encoded_range & 1)
					|| c_summ && encoded_range & 1)
				f |= bit2;			// set T, force Z zero
			else
				f |= bit3;			// no T, gamma may set Z
			if (encoded_range & 1)
				f |= bit0;			// N follows L35
			break;
	}

	if (prev_r || t_changes && f & bit2)
		f |= bit1;					// sticky R'ange flag

	return f;
}

uint16_t flags_calc(int i)
{
	int prev_n, prev_r, prev_t, prev_z, enc_ra, c_summ, p_summ, scheme;

	prev_n = i       &  1;
	prev_r = i >>  1 &  1;
	prev_t = i >>  2 &  1;
	prev_z = i >>  3 &  1;
	enc_ra = i >>  4 & 31;
	c_summ = i >>  9 &  1;
	p_summ = i >> 10 &  1;
	scheme = i >> 11 & 15;

	return flags_operation(prev_n, prev_r, prev_t, prev_z, enc_ra,
		c_summ, p_summ, scheme);
}

/* FOOTNOTES

Z:	As of this point, Z will be clear (output line is HIGH because Z
	is active low). If the incoming line is LOW, incoming Z is set,
	so we have its XOR swap accordingly.

*/
