/* gamma SRAM definition

INPUT BITS

	0-5		left input
	6-11	right input
	12		carry decision from theta
	13-15	a0..a4 operation

OUTPUT BITS

	0-5		result
	6		set iff result is zero (for zero-detect NAND)
	7		copy of gamma35 in subword 5 (reduces capacitance)

*/

#include "vm.h"

/* gamma5 is double-size, leading to odd spec in C */
uint8_t gamma[6][1 << 16];

int gamma_operation(int slot, int carry, int subword, int l, int r)
{
	int g = 0;
	int name = gamma_slots[slot + carry];

	if (l<0 || r<0 || l>c63 || r > c63 || subword < 0 || subword > c5)
		oops("gamma_operation");

	/* see "gamma operations" table from PDF */
	switch (name) {
		case '?':
			g = carry ? r : l;
			break;

		case '+':
			g = carry ? c63 & l + 1 : l;
			break;

		case 'k':
			g = carry ? 0 : l;
			break;

		case 'w':
			/* TODO stacked unary */
			break;

		case 's':
			g = sbox_unquote(gamma_sbox[subword][r][l]);
			break;

		case 'S':
			g = sbox_invert(alpha_sbox[subword][r], l);
			break;

		case 'u':
			/* TODO simple unary */
			break;

		case '^':
			g = c63 & (l ^ r);
			break;

		case 'l':
			g = rotate_tribble_left(l, r % 6);
			break;

		case 'L':
			/* Like 'l', but extends - sign from right. */
			g = rotate_tribble_left(l, r % 6);
			if (!r)	
				break;			/* left shift of 0 bits: identity */
			if (r > 35)
				r = 0;			/* left shift more than 35 bits: all 1s */
			g |= 63 & ((~((1L << r) - 1) & mask36) >> 6 * subword);
			break;

		case 'p':
			g = permute(r, l);
			break;

		case ' ':
			g = 0;				/* unallocated */
			break;

		default:
			oops("gamma name %c", name);
	}

	if (!g)
		/* zero flag */
		g |= g_zero_bit;

	if (subword == 5 && g & bit5)	// copy to reduce capacitance at N mux
		g |= bit7;

	return g;
}

uint16_t gamma_calc(int subword, int i)
{
	int l, r, slot, carry;

	if (use_ram)
		return gamma[subword][i];

	l = i & 63;
	r = i >> 6 & 63;
	carry = i >> 12 & 1;
	slot = 2 * (i >> 13 & 7);

	return gamma_operation(slot, carry, subword, l, r);
}

/* create entire gamma RAM */
void setup_gamma(void)
{
	static int done;
	int i, sw;

	if (done++)
		oops("double setup_gamma");

	/* see "SRAM input bit assignments" table from PDF */
	for (sw=0; sw<c6; sw++)
		for (i = 0; i < 1<<16; i++)
			gamma[sw][i] = gamma_calc(sw, i);
}

/* simulate the gamma layer */
long gamma_sim(int operation, int carries, long l, long r, int *zeros)
{
	int subword, c, q, z = 0, lt, rt;
	long res = 0;

	operation <<= 1 + c12;

	for (subword=0; subword<c6; subword++) {
		lt = c63 & l >> c6 * subword;
		rt = c63 & r >> c6 * subword;
		c = carries & 1 << subword ? 1 << c12 : 0;

		q = gamma_calc(subword, operation | c | rt << 6 | lt);

		res |= (long) (q & c63) << c6 * subword;
		ifnot (q & g_zero_bit)
			z |= 1 << subword;	// zero signals intended to be active-low
	}

	*zeros = z;

	return res;
}
