#include "vm.h"

// In this notation, 0 means the leftmost position, and 5 means rightmost.
static char *pset[] = {
	"012345", "123450", "234501", "345012", "450123", "501234", "432105",
	"321054", "210543", "105432", "054321", "543210", "102345", "210345",
	"312045", "412305", "512340", "021345", "032145", "042315", "052341",
	"013245", "014325", "015342", "012435", "012543", "012354", "452301",
	"014523", "230145", "024135", "031425", "031524", "142035", "253140",
	"304251", "415302", "520413", "035124", "140253", "251304", "302415",
	"413520", "524031", "102354", "315042", "542310", "513240", "021435",
	"034125", "043215", "103254", "240513", "453201", "521430", "534120",
	"120534", "201453", "034512", "035421", "105243", "130542", "254310",
	"510432",		// XXX 013 is reverse bits
};

// Apply 'perm' (0-63 from pset) to 6-bit integer 'i'.
int permute(int perm, int i)
{
	char *p;
	int pv, spv, out = 0;

	if (perm < 0 || perm > c63)
		oops("permute");
	p = pset[perm];

	// loop through place values of destination bits
	for (pv = c5; pv >= 0; --pv) {
		// compute the source place value
		spv = '0' + c5 - p[c5 - pv];
		// set if appropriate
		if (i & 1 << spv)
			out |= 1 << pv;
	}

	return out;
}

// FROM HERE DOWN, THIS IS ALL CODE FOR TESTING.

// In this notation, 0 means the rightmost octal digit, and 5 the leftmost.
static struct perms_of_6 {
	int p;				// permutation as 18-bit swizzle
	int p0;				// first of two steps; index into first 64 here
	int p1;				// second of two steps; identity (0) if one step
} permsteps[720];

static int nsteps;				// will reach 720

// Convert permutation "012345" to octal swizzle 543210.
// Note the change in numbering convention.
static int pset_to_swiz(char *p)
{
	int i = 0;

	for (i=0; *p; ++p)
		i = i << 3 | '5' - *p;
	return i;
}

// compose two octal swizzles (used as permutations)
// 054321 (right rotate 1 place) then 012345 (reflect) is equivalent to 123450.
static int compose_perm(int first, int second)
{
	int res = 0, i, q;

	for (i=0; i<6; i++)
		q = 7 & (second >> 3 * i),		// which octal digit position in first
		q = 7 & (first >> 3 * q),		// the digit itself
		res |= q << 3 * i;				// write to result

	return res;
}

// Do we know how to do permutation p already?
static struct perms_of_6 *find_way(int p)
{
	int i;

	for (i=0; i<nsteps; i++)
		if (permsteps[i].p == p)
			return permsteps + i;
	return NULL;
}

// Add a "discovery" as to how to do a permutation.
static void add_discovery(int p, int p0, int p1)
{
	struct perms_of_6 *way = find_way(p);

	if (way) return;
	if (nsteps >= 720)	// impossible
		oops("add_discovery");

	permsteps[nsteps].p = p;
	permsteps[nsteps].p0 = p0;
	permsteps[nsteps++].p1 = p1;
}

// Find fastest way to do each of 6! permutations. Run time 1.1 is ms.
static void learn_720(void)
{
	int i, j;

	// 64 base cases
	for (nsteps=0; nsteps<sizeof(pset)/sizeof(*pset); ++nsteps)
		permsteps[nsteps].p = pset_to_swiz(pset[nsteps]),
		permsteps[nsteps].p0 = nsteps,
		permsteps[nsteps].p1 = 0;

	if (nsteps != 64)
		oops("learn_720(a) %i", nsteps);

	for (i=0; i<64; i++)
		for (j=0; j<64; j++)
			add_discovery(compose_perm(permsteps[i].p, permsteps[j].p), i, j);

	if (nsteps != 720)
		oops("learn_720(b) %i", nsteps);
}

// "Octally" permute an array of six integers. Pointers may NOT overlap.
static void perm_six_ints(int octal_perm, int *in, int *out)
{
	int i;

	for (i=0; i<6; i++)
		out[i] = in[7 & (octal_perm >> 3 * i)];
}

// Create a permutation of 36 objects. Supply an integer array of at least 36.
// Randomized output will be SLIGHTLY imbalanced; this is for testing.
// Interpretation: bit 0 moves to bit p[0], bit 1 moves to bit p[1], etc.
static void random_perm_36(int *p)
{
	int pick, t;
	ulong r, i;

	for (i=0; i<36; i++)
		p[i] = i;

	urandom(sizeof(r), &r);
	for (i=36; i>1; i--) {
		pick = r % i;
		r /= i;
		if (i == 26 || i == 16)			// need more entropy
			urandom(sizeof(r), &r);
		t = p[i-1];						// swap selection with end
		p[i-1] = p[pick];
		p[pick] = t;
	}
}

static void transpose_perm_36(int *in, int *out)
{
	int i, r, c;

	for (i=0; i<36; i++)
		r = i/6, c = i%6, out[6*r+c] = in[6*c+r];
}

static void copy_perm_6(int *in, int *out)
{
	memcpy(out, in, 6 * sizeof(*in));
}

static long apply_perm_36(long in, int *p)
{
	long out = 0;
	int i;

	for (i=0; i<36; i++)
		if (in & 1L << i)
			out |= 1L << p[i];
	return out;
}

static void draw_perm_36(char *before, int *p, char *after)
{
	int i;

	if (before) printf("%s", before);
	for (i=35; i>=0; i--) {
		printf("\x1b[%im%i%i\x1b[m ", 31 + p[i] / 6, p[i] / 6, p[i] % 6);
		if (!i)
			printf("*");
		if (i % 6 == 0)
			printf("\n");
	}
	if (after) printf("%s", after);
}

// returns nonzero if a solution is found
static int solve_perm_36_step_1(
	int *p,				// 36 integers
	int r,				// row number: 0, 1, ..., 5
	int slow,			// limit is 720 if slow else 64
	long *x,			// PIT or PAT
	long *y				// PIT or PAT, but 0 if limit 64
) {
	int *row = 6 * r + p;
	int orig_row[6];
	int try;			// loop through the 64 one-CPU-cycle perms
	int octal;			// a row permutation to try
	int v;				// a row number being confirmed
	int c;				// a column number
	int limit;			// how many perms to try

	// Just starting? Initialize outputs.
	if (!r)
		*x = 0, *y = 0;

	// Store row we started with so we can iterate perms over it.
	copy_perm_6(row, orig_row);

	// Try the authorized perms.
	limit = slow ? 720 : 64;
	for (try=0; try<limit; ++try) {

		// Do the current perm.
		octal = permsteps[try].p;
		perm_six_ints(octal, orig_row, row);

		// see if any conflict with rows about this
		for (v=0; v<r; v++)
			for (c=0; c<6; c++)
				if (row[c] / 6 == p[6*v+c] / 6)
					goto nope;

		// Remember how we did this.
		*x = *x & ~(63L << 6 * r) | (long) permsteps[try].p0 << 6 * r;
		*y = *y & ~(63L << 6 * r) | (long) permsteps[try].p1 << 6 * r;

		// If this is the last row, puzzle is solved..
		if (r == 5) return 1;

		// If we can solve everything below this, puzzle is solved.
		if (solve_perm_36_step_1(p, r+1, slow, x, y))
			return 1;

		// Puzzle is not solved, but we have more quick perms to try.
nope:;
	}

	// Puzzle is not solved, and we are out of quick perms to try.
	copy_perm_6(orig_row, row);
	return 0;
}

// True iff row of 6 ints is in ascending order.
static int row_ascending(int *p)
{
	int i;

	for (i=0; i<5; i++)
		if (p[i] >= p[i+1])
			return 0;
	return 1;
}

// Always succeeds. Must put rows of p in ascending order.
// X and Y become arguments to PIT PIT or PAT PAT.
static void solve_perm_36_step_2(int *p, long *x, long *y)
{
	int row[6];					// a row of integers
	int r;						// a row number
	int try;					// 0, 1, ..., 719
	int octal;					// a permutation we're trying

	*x = 0;						// nothing yet
	*y = 0;

	for (r=0; r<6; r++) {		// do all rows

		// Want row to be ascending. Only one permutation works.
		for (try=0; try<720; ++try) {
			octal = permsteps[try].p;
			perm_six_ints(octal, p + 6 * r, row);
			if (row_ascending(row))
				goto found;
		}
		oops("solve_perm_36_step_2");	// Impossible scenario.

found:	copy_perm_6(row, p + 6 * r);
		*x |= (long) permsteps[try].p0 << 6 * r;
		*y |= (long) permsteps[try].p1 << 6 * r;
	}
}

void permutation_test_vector(struct permutation_test_case *ptc)
{
	int p[36], t[36];
	int watch = ptc -> watch;

	// initialize
	learn_720();

	// select problem
	random_perm_36(p);			// XXX Here's where to insert desired perm36.
	ptc -> expected_output = apply_perm_36(ptc -> left_operand, p);

	if (watch) {
		draw_perm_36("problem:\n", p, "\n");
	}

	// step 1
	ifnot (solve_perm_36_step_1(p, 0, 0, &ptc -> pit1, &ptc -> pit2)
		|| solve_perm_36_step_1(p, 0, 1, &ptc -> pit1, &ptc -> pit2)) {
		puts("no solution at step 1");
		return;
	}

	if (watch) {
		printf("step 1 solution uses PIT %012lo PIT %012lo:\n",
			ptc -> pit1, ptc -> pit2);
		draw_perm_36(NULL, p, "\n");
	}

	// step 2
	transpose_perm_36(p, t);
	solve_perm_36_step_2(t, &ptc -> pat3, &ptc -> pat4);
	transpose_perm_36(t, p);

	if (watch) {
		printf("step 2 solution uses PAT %012lo PAT %012lo:\n",
			ptc -> pat3, ptc -> pat4);
		draw_perm_36(NULL, p, "\n");
	}

	// step 3 is same as step 2, no transposition
	solve_perm_36_step_2(p, &ptc -> pit5, &ptc -> pit6);

	if (watch) {
		printf("step 3 solution uses PIT %012lo PIT %012lo:\n",
			ptc -> pit5, ptc -> pit6);
		draw_perm_36(NULL, p, "\n");
	}
}

void test_perms(void)
{
	struct permutation_test_case ptc = { 0 };

	urandom(sizeof(ptc.left_operand), &ptc.left_operand);
	ptc.left_operand &= mask36;

	ptc.watch = 1;
	permutation_test_vector(&ptc);

}
