#include "vm.h"
#include "stdarg.h"
#include "errno.h"

// TODO expand testing to other things that are working.

char covered[oc_NO_MORE];

/*
 *  Generic test case definition.
 *	Variable 'e' is chosen because it's short and ordinarily available.
 */
static struct stage {
	long l, r;								// operands
	long want, got;							// ALU result
	int opcode;								// that which we test
	int n_in, r_in, t_in, z_in;				// flags going in to the test
	int want_n, want_r, want_t, want_z;		// correct flags, all 0 or 1
	int want_interrupt;						// 0 or 1 for BOUND
	int got_n, got_r, got_t, got_z;			// observed flags
} e;

/*
 *	Execute a test function some number of times for random inputs.
 *	Verify that the expected output occurred and report any disparity.
 */
void exercise(char *section, int (*setup_fn)(int *), int reps)
{
	int rep;	// iteration
	int gen;	// local storage for setup_fn
	int flags, interrupt_flag;

	printf("testing %s\n", section);

	for (rep = 0; rep < reps; ++rep) {

		// select left and right operands at random
		urandom(sizeof(e.l), &e.l);
		e.l &= mask36;
		urandom(sizeof(e.r), &e.r);
		e.r &= mask36;

		// select flag inputs at random
		{
			int flags;
			urandom(sizeof(flags), &flags);
			e.n_in = !(flags & 1);
			e.r_in = !(flags & 2);
			e.t_in = !(flags & 4);
			e.z_in = !(flags & 8);
		}

		// successively set up tests for these inputs
		gen = -1;
		while (setup_fn(&gen)) {

			// feed this test case through the ALU
B:			flags = bit0 & e.n_in | bit1 & e.r_in << 1 | 
					bit2 & e.t_in << 2 | bit3 & ~e.z_in << 3;
			e.got = alu(e.opcode, e.l, e.r, &flags);
			covered[e.opcode] = 1;
			e.got_n = !!(flags & bit0);
			e.got_r = !!(flags & bit1);
			e.got_t = !!(flags & bit2);
			e.got_z =  !(flags & bit3);
			interrupt_flag = !!(flags & bit4);

			if (e.got > mask36 || e.got < 0)
				oops("ALU output too many bits: %li\n", e.got);

			if (e.want == e.got && e.want_n == e.got_n && e.want_r == e.got_r
				&& e.want_t == e.got_t && e.want_z == e.got_z 
				&& e.want_interrupt == interrupt_flag)
				continue;

			// report failure if we reach here
			printf("FAIL TEST:\n");
			printf("    input: %li %s %li %c %c %c %c (L pole %s)\n",
				e.l, Decoder[e.opcode].name, e.r,
				e.n_in ? 'N' : 'n', e.r_in ? 'R' : 'r',
				e.t_in ? 'T' : 't', e.z_in ? 'Z' : 'z',
				e.l & pole ? "set" : "clear");
			if (e.l == 010101010101L * (e.l & 63))
				printf("        l: {%li}\n", e.l & 63);
			if (e.r == 010101010101L * (e.r & 63))
				printf("        r: {%li}\n", e.r & 63);
			printf("   result: %li %c %c %c %c %c\n", e.got, 
				interrupt_flag ? 'I' : 'i',	e.got_n ? 'N' : 'n',
				e.got_r ? 'R' : 'r',
				e.got_t ? 'T' : 't', e.got_z ? 'Z' : 'z');
			printf("   wanted: %li %c %c %c %c %c\n", e.want,
				e.want_interrupt ? 'I' : 'i', e.want_n ? 'N' : 'n',
				e.want_r ? 'R' : 'r',
				e.want_t ? 'T' : 't', e.want_z ? 'Z' : 'z');
			exit(-1);
		}
	}
}

/*
 *	Skew random 36-bit words to be near discontinuities.
 *	This emphasizes tests just above and below 0 and pole.
 *	There are exactly 32 possible results from this function:
 *	The first and last two bits are random, and the middle bits
 *	are all 1s or all 0s.
 */
static long polar_skew(long w)
{
	long z = w & 0600000000003;

	if (w & 0000000100000L)
		z |= 0177777777774;

	return z;
}

/*
 *	Set up a single test for the stated arithmetic.
 *	This template instance must fill in: e.want, e.want_n, e.want_t, e.want_z
 *
 *	'arith' is one of [+ - m] for add, subtract, reverse subtract.
 *	'signedness' bits are 4 (left), 2 (right), 1 (result); set means signed.
 *	'pc' is true iff the previous carry is added or subtracted.
 *	'jc' (just compute) prevents changing T (MIN, MAX, BOUND),
 *	also Z (BOUND) and N (BOUND).
 */
void setup_additive_test(int arith, int signedness, int pc, int jc)
{
	long l = e.l, r = e.r, sum = 0, clip;

	// Use signedness info to extent l and r to true 64-bit values.
	if (signedness & 4 && l & pole)
		l |= comp36;
	if (signedness & 2 && r & pole)
		r |= comp36;

	// Compute sum with, if applicable, previous carry.
	pc = pc ? e.t_in : 0;
	switch (arith) {
		case '+': sum = l + r + pc; break;
		case '-': sum = l - r - pc; break;
		case 'm': sum = r - l - pc; break;
	}

	// N, Z set iff the /real/ sum is negative or zero.
	e.want_n = sum < 0;
	e.want_z = !sum;

	// Clip sum to 36 bits. That's what the ALU should give us.
	clip = sum & mask36;
	e.want = clip;

	// Re-extend sign if applicable.
	if (signedness & 1 && clip & pole)
		clip |= comp36;

	// T flag (overflow) <==> clip does not match sum.
	e.want_t = clip != sum;

	// Array boundary overflow? Want I, don't touch Z or N.
	if (jc == 2) {
		e.want_interrupt = !e.want_n;
		e.want_n = e.n_in;
		e.want_z = e.z_in;
	}

	// Comparisons don't set T (or R).
	if (jc)
		e.want_t = e.t_in;

	e.want_r = e.r_in || !jc && e.want_t;
}

/*
 *	Generate successive additive and comparison tests.
 *	*gen is available as a local variable and is -1 on the first pass.
 *	Returns zero iff all tests have been run.
 */
int setup_additive_tests(int *gen)
{
	static int iterations;

	static struct add_cases {
		int opcode;
		int arith;
		int signedness;
		int pc;
		int just_compare;
	} cases[] = {
		oc_A_uuu, '+', 0, 0, 0, oc_A_uus, '+', 1, 0, 0,
		oc_A_usu, '+', 2, 0, 0, oc_A_uss, '+', 3, 0, 0,
		oc_A_suu, '+', 4, 0, 0, oc_A_sus, '+', 5, 0, 0,
		oc_A_ssu, '+', 6, 0, 0, oc_A_sss, '+', 7, 0, 0,
		oc_S_uuu, '-', 0, 0, 0, oc_S_uus, '-', 1, 0, 0,
		oc_S_usu, '-', 2, 0, 0, oc_S_uss, '-', 3, 0, 0,
		oc_S_suu, '-', 4, 0, 0, oc_S_sus, '-', 5, 0, 0,
		oc_S_ssu, '-', 6, 0, 0, oc_S_sss, '-', 7, 0, 0,
		oc_RS_uuu, 'm', 0, 0, 0, oc_RS_uus, 'm', 1, 0, 0,
		oc_RS_usu, 'm', 2, 0, 0, oc_RS_uss, 'm', 3, 0, 0,
		oc_RS_suu, 'm', 4, 0, 0, oc_RS_sus, 'm', 5, 0, 0,
		oc_RS_ssu, 'm', 6, 0, 0, oc_RS_sss, 'm', 7, 0, 0,
		oc_AC_uuu, '+', 0, 1, 0, oc_AC_uus, '+', 1, 1, 0,
		oc_AC_usu, '+', 2, 1, 0, oc_AC_uss, '+', 3, 1, 0,
		oc_AC_suu, '+', 4, 1, 0, oc_AC_sus, '+', 5, 1, 0,
		oc_AC_ssu, '+', 6, 1, 0, oc_AC_sss, '+', 7, 1, 0,
		oc_SC_uuu, '-', 0, 1, 0, oc_SC_uus, '-', 1, 1, 0,
		oc_SC_usu, '-', 2, 1, 0, oc_SC_uss, '-', 3, 1, 0,
		oc_SC_suu, '-', 4, 1, 0, oc_SC_sus, '-', 5, 1, 0,
		oc_SC_ssu, '-', 6, 1, 0, oc_SC_sss, '-', 7, 1, 0,
		oc_RSC_uuu, 'm', 0, 1, 0, oc_RSC_uus, 'm', 1, 1, 0,
		oc_RSC_usu, 'm', 2, 1, 0, oc_RSC_uss, 'm', 3, 1, 0,
		oc_RSC_suu, 'm', 4, 1, 0, oc_RSC_sus, 'm', 5, 1, 0,
		oc_RSC_ssu, 'm', 6, 1, 0, oc_RSC_sss, 'm', 7, 1, 0,
		oc_CMP_uu, '-', 0, 0, 1, oc_CMP_us, '-', 2, 0, 1,
		oc_CMP_su, '-', 4, 0, 1, oc_CMP_ss, '-', 6, 0, 1,
		oc_BOUND_s, '-', 2, 0, 2, oc_BOUND_u, '-', 0, 0, 2, 
	};

	if (*gen < 0)		// first pass? initialize counter
		*gen = 0;
	if (*gen >= sizeof(cases) / sizeof(*cases))
		return 0;		// all tests have run

	// Target enough tests to areas near wraparound boundaries.
	if (iterations++ < 65536 && (iterations & 15) == 0) {
		e.l = polar_skew(e.l);
		e.r = polar_skew(e.r);
	}

	e.opcode = cases[*gen].opcode;
	e.want_interrupt = 0;
	setup_additive_test(cases[*gen].arith, 
			cases[*gen].signedness, cases[*gen].pc, cases[*gen].just_compare);
	++*gen;

	return 1;
}

/*
 *	Set up a single test for a bitwise boolean function.
 *	This template instance must fill in: e.want, e.want_n, e.want_t, e.want_z
 *
 *	'l_op_r' contains result bits for l,r = 0,0 0,1 1,0 1,1 as bits 1, 2, 4, 8.
 */
void setup_bitwise_test(int l_op_r)
{
	int which, i;
	long res = 0;

	if (l_op_r == 15)		// IGT has a requirement on the right argument
		e.r = mask36;
	if (l_op_r == 5)		// NR has a requirement on the left argument
		e.l = mask36;

	for (i=0; i<36; i++) {
		which = e.l & 1L << i ? 4 : 1;
		which <<= e.r & 1L << i ? 1 : 0;
		if (which & l_op_r)
			res |= 1L << i;
	}

	e.want = res;
	e.want_t = e.t_in;		// Correct semantic. Bitwise isn't arithmetic.
	e.want_n = !!(res & pole);	// New semantic: can alter N flag.
	e.want_z = !res;
	e.want_r = e.r_in;
}

/*
 *	Generate successive bitwise boolean tests.
 *	*gen is available as a local variable and is -1 on the first pass.
 *	Returns zero iff all tests have been run.
 */
int setup_bitwise_tests(int *gen)
{
	static struct add_cases {
		int opcode;
		int l_op_r;
	} cases[] = {
		oc_IGF,   0, oc_IGT,  15, oc_XL,   12, oc_XR,   10,
		oc_NL,    3, oc_NR,    5, oc_AND,   8, oc_NAND,  7,
		oc_OR,   14, oc_NOR,   1, oc_XOR,   6, oc_XNOR,  9,
		oc_LANR,  4, oc_RANL,  2, oc_LONR, 13, oc_RONL, 11,
	};

	if (*gen < 0)		// first pass? initialize counter
		*gen = 0;
	if (*gen >= sizeof(cases) / sizeof(*cases))
		return 0;		// all tests have run

	e.opcode = cases[*gen].opcode;
	e.want_interrupt = 0;
	setup_bitwise_test(cases[*gen].l_op_r);
	++*gen;

	return 1;
}

/*
 *	Set up a single test for MIN / MAX.
 *	This template instance must fill in: e.want, e.want_n, e.want_t, e.want_z
 *
 *	'arith' is one of [n x] for min, max.
 *	'signedness' bits are 4 (left), 2 (right), 1 (result); set means signed.
 */
void setup_min_max_test(int arith, int signedness)
{
	long l = e.l, r = e.r, out = 0, clip;

	// Use signedness info to extent l and r to true 64-bit values.
	if (signedness & 4 && l & pole)
		l |= comp36;
	if (signedness & 2 && r & pole)
		r |= comp36;

	// Find minimum or maximum.
	switch (arith) {
		case 'n': out = l < r ? l : r; break;
		case 'x': out = l > r ? l : r; break;
	}

	// N, Z set iff the /real/ outcome is negative or zero.
	e.want_n = out < 0;
	e.want_z = !out;

	// Clip sum to 36 bits. That's what the ALU should give us.
	clip = out & mask36;
	e.want = clip;

	// Re-extend sign if applicable.
	if (signedness & 1 && clip & pole)
		clip |= comp36;

	// T flag (overflow) <==> clip does not match out.
	e.want_t = clip != out;
	e.want_r = e.r_in || e.want_t;
}

/*
 *	Generate successive MIN / MAX tests.
 *	*gen is available as a local variable and is -1 on the first pass.
 *	Returns zero iff all tests have been run.
 */
int setup_min_max_tests(int *gen)
{
	static int iterations;

	static struct add_cases {
		int opcode;
		int arith;
		int signedness;
	} cases[] = {
		oc_MIN_uuu, 'n', 0, oc_MIN_uus, 'n', 1,
		oc_MIN_usu, 'n', 2, oc_MIN_uss, 'n', 3,
		oc_MIN_suu, 'n', 4, oc_MIN_sus, 'n', 5,
		oc_MIN_ssu, 'n', 6, oc_MIN_sss, 'n', 7,
		oc_MAX_uuu, 'x', 0, oc_MAX_uus, 'x', 1,
		oc_MAX_usu, 'x', 2, oc_MAX_uss, 'x', 3,
		oc_MAX_suu, 'x', 4, oc_MAX_sus, 'x', 5,
		oc_MAX_ssu, 'x', 6, oc_MAX_sss, 'x', 7,
	};

	if (*gen < 0)		// first pass? initialize counter
		*gen = 0;
	if (*gen >= sizeof(cases) / sizeof(*cases))
		return 0;		// all tests have run

	// Target enough tests to areas near wraparound boundaries.
	if (iterations++ < 65536 && (iterations & 15) == 0) {
		e.l = polar_skew(e.l);
		e.r = polar_skew(e.r);
	}

	e.opcode = cases[*gen].opcode;
	e.want_interrupt = 0;
	setup_min_max_test(cases[*gen].arith, cases[*gen].signedness);
	++*gen;

	return 1;
}

/*
 *	Set up a single test for crypto/hash mix and unmix.
 *	This template instance must fill in: e.want, e.want_n, e.want_t, e.want_z
 *
 *	'direction' is 'm' for MIX or 'x' for XIM.
 *
 *	XXX All we can really show today is that MIX and XIM are inverses.
 *	    There is an infelicity that this setup modifies the requested test,
 *	    because we have no independent code that can mimic it.
 */
void setup_mix_xim_test(int direction)
{
	int inverse;

	inverse = direction == 'm' ? oc_XIM : oc_MIX;
	e.want = e.l;
A:	e.l = alu(inverse, e.l, e.r, NULL);
	covered[inverse] = 1;

	e.want_t = e.t_in;		// MIX is not arithmetic.
	e.want_n = !!(e.want & pole);
	e.want_z = !e.want;
	e.want_r = e.r_in;
}

/*
 *	Generate successive crypto tests.
 *	*gen is available as a local variable and is -1 on the first pass.
 *	Returns zero iff all tests have been run.
 */
int setup_mix_xim_tests(int *gen)
{
	static struct add_cases {
		int opcode;
		int direction;
	} cases[] = { oc_MIX, 'm', oc_XIM, 'x' };

	if (*gen < 0)		// first pass? initialize counter
		*gen = 0;
	if (*gen >= sizeof(cases) / sizeof(*cases))
		return 0;		// all tests have run

	e.opcode = cases[*gen].opcode;
	e.want_interrupt = 0;
	setup_mix_xim_test(cases[*gen].direction);
	++*gen;

	return 1;
}

/*
 *	Set up a rotate test.
 *	This template instance must fill in: e.want, e.want_n, e.want_t, e.want_z
 *
 *	Let's talk about gcc and x86: shift amounts only work to one minus the
 *	word size, and then they wrap, which is totally wrong. So for 'L' and 'R',
 *	there are some & 63 operations needed for portability, but not needed
 *	on the buggy platform I am using.
 */
void setup_rotate_test(int mode, int signedness)
{
	int lp;					// left positions
	int rp;
	long w, x;

	switch (mode) {
		case 'l':			// rotate left
			lp = (e.r & 63) % 36;
			e.want = mask36 & (e.l << lp | e.l >> 36 - lp);

			e.want_t = e.t_in;		// Rotate is not arithmetic.
			e.want_n = !!(e.want & pole);
			e.want_z = !e.want;
			e.want_r = e.r_in;
			break;

		case 'L':			// logical shift left
			lp = e.r & 63;
			e.want = mask36 & e.l << lp;

			e.want_t = e.t_in;		// Logical shift is not arithmetic.
			e.want_n = !!(e.want & pole);
			e.want_z = !e.want;
			e.want_r = e.r_in;
			break;

		case 'R':			// logical shift right
			/*
			 *	Shifts right are implemented via left rotations in order
			 *	to save a slot in gamma. We don't offer right rotations.
			 *	A right shift of 1 is coded like a left shift of 35.
			 *	Any right shift >= 36 is coded like a left shift of 36.
			 */
			lp = e.r & 63;
			if (lp > 35) {
				e.want = 0;
				e.want_z = !e.want;
				break;
			}
			else if (!lp)	// For portability on non-MWA 36-bit CPUs,
				lp = 36;	// we won't shift by their word size.
			e.want = mask36 & e.l >> 36 - lp;

			e.want_t = e.t_in;		// Logical shift is not arithmetic.
			e.want_n = !!(e.want & pole);
			e.want_z = !e.want;
			e.want_r = e.r_in;
			break;

		case '}':			// arithmetic shift right
			lp = e.r & 63;
			if (lp > 35)
				rp = 36;	// full loss of shifted word except sign
			else if (!lp)
				rp = 0;
			else
				rp = 36 - lp;
			w = e.l;		// input word
			if (signedness & 2 && w & pole)
				w |= comp36;
			w >>= rp;
			e.want = w & mask36;

			e.want_t = signedness & 1 ? w >= pole : w < 0;
			e.want_n = signedness & 2 ? !!(e.l & pole) : 0;
			e.want_z = !e.want;		// Can only with non-negative inputs.
			e.want_r = e.r_in || e.want_t;
			break;

		case '{':			// arithmetic shift left
			lp = e.r & 63;

			w = e.l;			// input word
			if (signedness & 2 && w & pole)
				w |= comp36;	// sign-extended input
			e.want = mask36 & w << lp;
			x = e.want;			// output word
			if (signedness & 1 && x & pole)
				x |= comp36;	// needs sign extension
			x >>= lp;			// undo shift
			e.want_t = w != x;	// overflow if input unrecoverable
			e.want_n = w < 0;
			e.want_z = !w;
			e.want_r = e.r_in || e.want_t;
			break;
	}
}

/*
 *	Generate successive crypto tests.
 *	*gen is available as a local variable and is -1 on the first pass.
 *	Returns zero iff all tests have been run.
 */
int setup_rotate_tests(int *gen)
{
	static int iterations;

	static struct add_cases {
		int opcode;
		int mode;
		int signedness;
	} cases[] = {
		oc_ROL, 'l', 0, oc_LSL, 'L', 0, oc_LSR, 'R', 0,
		oc_ASL_uu, '{', 0, oc_ASL_us, '{', 1,
		oc_ASL_su, '{', 2, oc_ASL_ss, '{', 3,
		oc_ASR_uu, '}', 0, oc_ASR_us, '}', 1,
		oc_ASR_su, '}', 2, oc_ASR_ss, '}', 3,
	};

	if (*gen < 0) {							// first pass? init counter
		*gen = 0;
		e.r = 010101010101 * (e.r & 63);	// swizzle rotate amount
	}

	if (*gen >= sizeof(cases) / sizeof(*cases))
		return 0;		// all tests have run

	// Target enough tests to left operands near wraparound boundaries.
	if (iterations++ < 65536 && (iterations & 15) == 0)
		e.l = polar_skew(e.l);

	e.opcode = cases[*gen].opcode;
	e.want_interrupt = 0;
	setup_rotate_test(cases[*gen].mode, cases[*gen].signedness);
	++*gen;

	return 1;
}

/*
 *	Set up a single test for multiplication support.
 *	This template instance must fill in: e.want, e.want_n, e.want_t, e.want_z
 */
void setup_mul_test(void)
{
	long six = e.r % 6, sixes = six * 010101010101L;
	int i;

	e.want_t = 0;					// true except sometimes for MH
	e.want_n = 0;

	switch (e.opcode) {
		case oc_MH:
		case oc_MHNS:
			e.r = sixes;
			e.want = 0;
			for (i=0; i<6; i++)
				e.want |= (07700 & ((63 & e.l >> 6 * i) * (63 & e.r >> 6 * i)))
					>> 6 << 6 * i;
			if (e.opcode == oc_MHNS) break;
			e.want = 0777777777700 & (e.want << 6) | 077 & (e.want >> 30);
			if (e.want & 63)
				e.want_t = 1;		// MH overflow
			break;

		case oc_ML:
			e.r = sixes;
			e.want = 0;
			for (i=0; i<6; i++)
				e.want |= (63 & ((e.l >> 6 * i) * (e.r >> 6 * i))) << 6 * i;
			break;

		case oc_COPY:
			// XXX swizzling not tested here, only "well-formed" copies
			e.r = sixes;
			e.want = (63 & (e.l >> six * 6)) * 010101010101;
			break;

		case oc_DSL:
			e.want = mask36 & (e.l << 6 | e.r >> 30);
	}

	e.want_z = !e.want;
	e.want_r = e.r_in || e.want_t;
}

/*
 *	Generate successive multiplication support tests.
 *	*gen is available as a local variable and is -1 on the first pass.
 *	Returns zero iff all tests have been run.
 */
int setup_mul_tests(int *gen)
{
	static long right_operand;
	static struct mul_cases {
		int opcode;
	} cases[] = { oc_MH, oc_MHNS, oc_ML, oc_COPY, oc_DSL };

	if (*gen < 0)		// first pass? initialize counter
		right_operand = e.r, *gen = 0;
	if (*gen >= sizeof(cases) / sizeof(*cases))
		return 0;		// all tests have run

	e.opcode = cases[*gen].opcode;
	e.r = right_operand;
	e.want_interrupt = 0;
	setup_mul_test();
	++*gen;

	return 1;
}

/*
 *	Set up tests for permutations. PAIT is fast, but PIT (includes PAT)
 *	can only manage ~800/second, because the test uses full 36-bit permutes
 *	that must be decomposed into PIT and PAT instructions.
 *	This template instance must fill in: e.want, e.want_n, e.want_t, e.want_z
 */
void setup_perm_test(void)
{
	long x;
	struct permutation_test_case ptc = { 0 };

	switch (e.opcode) {

		case oc_PIT:
			ptc.left_operand = e.l;
			permutation_test_vector(&ptc);
			x = alu(oc_PIT, e.l, ptc.pit1, NULL);
			x = alu(oc_PIT, x, ptc.pit2, NULL);
			x = alu(oc_PAT, x, ptc.pat3, NULL);
			x = alu(oc_PAT, x, ptc.pat4, NULL);
			e.l = alu(oc_PIT, x, ptc.pit5, NULL);
			e.r = ptc.pit6;
			e.want = ptc.expected_output;
			covered[oc_PAT] = 1;
			break;

		case oc_PAIT:
			e.want = alu(oc_PIT, alu(oc_PAT, e.l, e.r, NULL), e.r, NULL);
			break;

		default:
			oops("setup_perm_test");
	}

	e.want_n = !!(e.want & pole);
	e.want_z = !e.want;
	e.want_t = e.t_in;
	e.want_r = e.r_in;
}

/*
 *	Generate successive permutation tests.
 *	*gen is available as a local variable and is -1 on the first pass.
 *	Returns zero iff all tests have been run.
 */
int setup_perm_tests(int *gen)
{
	static long left_operand, right_operand;
	static struct add_cases {
		int opcode;
	} cases[] = {
		oc_PAIT,
		oc_PIT,								// PAT is included in this test.
	};

	if (*gen < 0) {							// first pass? init counter
		*gen = 0;
		left_operand = e.l;
		right_operand = e.r;
	}

	if (*gen >= sizeof(cases) / sizeof(*cases))
		return 0;		// all tests have run

	e.opcode = cases[*gen].opcode;
	e.want_interrupt = 0;
	e.l = left_operand;
	e.r = right_operand;
	setup_perm_test();
	++*gen;

	return 1;
}

/*
 *	Test CRF and NUDGE.
 *	This template instance must fill in:
 *		e.want, e.want_n, e.want_r, e.want_t, e.want_z
 *	This setup modifies the right operand.
 */
void setup_misc_test(void)
{
	long modif, i, mask = 0;

	urandom(sizeof(modif), &modif);
	e.r &= modif % 37 - 1;				// uniformly 0 to 36 bits wide

	switch (e.opcode) {

		case oc_NUDGE:
			for (i=pole; i; i >>= 1)
				if (e.r & i) {
					mask = i - 1;
					break;
				}
			e.want = e.l & ~mask | e.r & mask;
			e.want_n = !!(e.want & pole);
			e.want_z = !e.want;
			e.want_t = e.t_in;
			e.want_r = e.r_in;
			break;

		case oc_CRF:					// CLR = Clear Range Flag
			e.want = 0;
			e.want_n = e.n_in;
			e.want_r = 0;
			e.want_t = e.t_in;
			e.want_z = e.z_in;
			break;

		default:
			oops("setup_misc_test");
	}
}

/*
 *	Generate successive NUDGE tests.
 *	*gen is available as a local variable and is -1 on the first pass.
 *	Returns zero iff all tests have been run.
 */
int setup_misc_tests(int *gen)
{
	static long right_op;
	static struct add_cases {
		int opcode;
	} cases[] = { oc_CRF, oc_NUDGE };

	if (*gen < 0)							// first pass? init counter
		right_op = e.r, *gen = 0;

	if (*gen >= sizeof(cases) / sizeof(*cases))
		return 0;		// all tests have run

	e.opcode = cases[*gen].opcode;
	e.r = right_op;
	e.want_interrupt = 0;
	setup_misc_test();
	++*gen;

	return 1;
}

void test_stuff(void)
{
	int i, sum = 0;

	exercise("additive", setup_additive_tests, 100000);
	exercise("bitwise", setup_bitwise_tests, 100000);
	exercise("min/max", setup_min_max_tests, 100000);
	exercise("rotate", setup_rotate_tests, 100000);
	exercise("multiply", setup_mul_tests, 100000);
	exercise("mix/xim", setup_mix_xim_tests, 100000);
	exercise("permute", setup_perm_tests, 1000);
	exercise("misc", setup_misc_tests, 100000);

	for (i=0; i<oc_NO_MORE; i++) {
		if (covered[i]) ++sum;
		else printf("  UNTESTED: %s\n", Decoder[i].name);
	}
	printf("Passed: %i opcodes\n", sum);
}

/* FOOTNOTES

A:	As we only compute the inverse here, there is nothing to do with
	flags in either direction.

B:	Remember that Z is active low.

*/
