/* theta SRAM definition

INPUT BITS

	0-5		carry bits from a0..a5, named c0..c5
	6-11	propagate bits from a0..a5, named p0..p5
	12		T flag (often adjusted)
	13-15	operation (often extended by T flag)

OUTPUT BITS

	0-5		carry decisions to g0..g5, named d0..d5
	6		"summary" of carries c0..c5
	7		"summary" of propagates p0..p5

*/

#include "vm.h"

uint8_t theta[1 << 16];

int theta_operation(int slot, int t_flag, int props, int carries)
{
	int th = 0, i;
	int name = theta_slots[slot + t_flag];

	if (props<0 || carries<0 || props>c63 || carries > c63 || t_flag & ~1)
		oops("theta_operation");

	/* see "theta operations" table from PDF */
	switch (name) {
		case '+':
			if (props & carries)
				break;		/* special case for integer absolute value */
			th = t_flag;	/* base case */
			for (i=2; i < c64; i <<= 1) {
				/* Propagate and carry is mutually exclusive here. */
				/* Carry from previous? Set carry decision. */
				th |= i & carries << 1;
				/* Propagate from previous? Set carry decision. */
				th |= i & (props & th) << 1;
			}
			if (th & bit5)
				th |= bit6;		// zeta's view of d5
			break;

		case '<':				// bit 6 muxes the encoded range
			th = props < carries ? bit6 | c63 : 0;
			break;

		case '>':				// bit 6 muxes the encoded range
			th = props > carries ? bit6 | c63 : 0;
			break;

		case '{':
			// TODO No customers here.
			break;

		case '}':
			// TODO NUDGE is only customer today, so this is very simple.
			for (i=1; i & 63; i <<= 1)
				if (carries & i)
					th |= i-1;
			break;

		case '0':				// already zero
			if (carries)
				th |= bit6;		// ASR encroachment by 0
			if (props)
				th |= bit7;		// ASR encroachment by 1
			break;

		case '1':
			th = c63;
			break;

		case 'n':			/* replicate c_5 to all d_i */
			if (carries & c32)
				th = c63;
			break;

		case ' ':			/* unallocated */
			th = 0;
			break;

		default:
			oops("theta name '%c' slot %i", name, slot);
	}

	return th;
}

uint16_t theta_calc(int i)
{
	int op, t_flag, props, carries;

	if (use_ram)
		return theta[i];

	carries = 63 & i;
	props = 63 & i >> 6;
	t_flag = 1 & i >> 12;
	op = 2 * (i >> 13);

	return theta_operation(op, t_flag, props, carries);
}

/* create entire theta RAM */
void setup_theta(void)
{
	static int done;
	int i;

	if (done++)
		oops("double setup_theta");

	/* see "SRAM input bit assignments" table from PDF */
	for (i=0; i < 1 << 16; i++)
		theta[i] = theta_calc(i);
}

/* simulate the theta RAM */
int theta_sim(int operation, int t_flag, int props, int carries)
{
	return theta_calc
		(operation << 1 + c12 | t_flag << c12 | props << c6 | carries);
}
