#include "vm.h"
#include "ctype.h"
#include "stdarg.h"
#include "errno.h"
#include "unistd.h"

int show = 0;
int tracing;
int use_ram;		/* must not set before RAM initialized */

const long			/* rather enum, but promotion semantics are horrible */
	pole = 1L << 35,
	two36 = 1L << 36,
	mask36 = (1L << 36) - 1,
	comp36 = ~((1L << 36) - 1);

/*
	oops() terminates the program with an error message.
	Output goes to standard error. You may format it to suit
	your needs. The exit code is 1.

	If the message begins with a semicolon, the semicolon
	is not printed, and the program does not exit.
*/
void oops(char *fmt, ...)
{
	va_list ap;
	int info = *fmt == ';';

	fflush(stdout);	// get last gasp of helpful text
	if (info)
		++fmt;
	va_start(ap, fmt); 
	vfprintf(stderr, fmt, ap);
	fprintf(stderr, "\n");
	va_end(ap);
	if (!info)
		exit(1);
}

/* convert ' ' thru 'a' to 0 thru 63; does not use " or \ */
int sbox_unquote(int c)
{
	/* back 'c' down so it may be " or \ */
	if (c > '"') --c;
	if (c > '\\'-1) --c;
	c -= 32;
	if (c < 0 || c > c63)
		oops("sbox_unquote %i", c);
	return c;
}

/* find input to quoted S-box that produces 'i' */
int sbox_invert(char *q, int i)
{
	char *s;

	i += 32;
	/* nudge 'i' up so it may not be " or \ */
	if (i >= '"') ++i;
	if (i >= '\\') ++i;
	ifnot (s = strchr(q, i))
		oops("sbox_invert %s [%i]", q, i);
	i = s - q;
	if (i < 0 || i > c63)
		oops("sbox_invert %i", i);
	return i;
}

/* set buffer from CSPRNG */
void urandom(int len, void *buf)
{
	static FILE *uFile;                     // /dev/urandom
	static char *path = "/dev/urandom";

	if (len < 1) return;                    // no "leftover" space to randomize

	if (!uFile && !(uFile = fopen(path, "r")))
		oops("cannot open CSPRNG %s", path);

	do {
		errno = 0;
		if (len == fread(buf, 1, len, uFile))
			return;                         // success (almost always the case)
	} while (errno == EINTR);

	oops("error reading CSPRNG %s", path);
}

/* convert opcode name to opcode via decoder table scan */
int find_opcode(char *s)
{
	int i;

	for (i = 0; i < oc_NO_MORE; i++)
		if (!strcmp(s, Decoder[i].name))
			return i;

	return -1;
}

/*
 *	Shuffle bits of word A according to base-36 template string.
 *	May expand without notice to additional inputs; code already present.
 *	Whitespace in format is ignored.
 *
 *	y = wire("3e0", x) places bits 3, 14, and 0 of x into bits 2, 1, 0 of y.
 *
 *	y = wire("ztnhb5ysmga4xrlf93wqke82vpjd71uoic60", x) does a beta-transpose.
 *
 *	If enabled, may put A, B, C into format string to switch inputs as of
 *	their positions.
 *
 *	Warning: This is not a fast function; actually /using/ the above transpose
 *	slows the whole ALU by 3.5 times. But it's ok for computing tables once.
 */
long slow_wire(char *fmt, long A)
{
	static long B, C;			/* possible future use */
	long in = A, y = 0;
	int c, next_bit = 35;

	while (c = *fmt++) {

		if (isdigit(c))
			c -= '0';
		else if (islower(c))
			c -= ('a' - 10);
		else
			goto non_position;

		if (in & 1L << c)
			y |= 1L << next_bit;
		next_bit -= 1;
		continue;

non_position:
		if (c == 'A') in = A;
		else if (c == 'B') in = B;
		else if (c == 'C') in = C;
		else if (isspace(c)) ;
		else oops("wire %c\n", c);
	}

	if (next_bit >= 0)
		y >>= next_bit + 1;
	return y;
}

/* rotate tribble left */
int rotate_tribble_left(int t, int n)
{
	n %= 6;
	t <<= n;
	return 63 & (t | t >> 6);
}

/* print tribble as bits */
void print6(char *prefix, uint t, char *suffix)
{
	uint i;

	if (prefix)
		printf("%s", prefix);

	for (i=bit5; i; i >>= 1)
		printf("%s", t & i ? "1" : "0");

	if (suffix)
		printf("%s", suffix);
}

/* print 36-bit word as bits */
void print36(char *prefix, ulong w, char *suffix)
{
	int i;

	if (prefix)
		printf("%s", prefix);

	for (i=5; i>=0; --i)
		print6(NULL, 63 & w >> 6 * i, " " + !i);

	if (suffix)
		printf("%s", suffix);
}

/* test code */
int main(int argc, char *argv[])
{
	char *e;
	long l, r;
	int opcode, flags = 0;

	nice(9999);
	if (sizeof(int) < 4 || sizeof(long) < 8)
		oops("unanticipated architecture");

	setup_stats();
	setup_decoder();
	setup_alu();

	if (getenv("ALU_CLEAR"))
		fprintf(stderr, "\x1b""c");

	switch (argc) {
		case 2:								// various MIX and XIM tests
			some_stats(argv[1]);
			break;

		case 5:
			e = argv[4];
			if (strchr(e, 'N')) flags |= bit0;
			if (strchr(e, 'R')) flags |= bit1;
			if (strchr(e, 'T')) flags |= bit2;
			if (strchr(e, 'Z')) flags |= bit3;
			nobreak;

		case 4:								// computation with ALU display
			opcode = find_opcode(argv[2]);
			if (opcode < 0)
				oops("invalid opcode");

			l = strtol(argv[1], &e, 0);
			ifnot ((l & comp36) == 0 || (l | mask36) == -1)
				oops("left argument out of range");
			if (*e)
				oops("left argument bad syntax");

			r = strtol(argv[3], &e, 0);
			ifnot ((r & comp36) == 0 || (r | mask36) == -1)
				oops("right argument out of range");
			if (*e)
				oops("right argument bad syntax");

			if (getenv("ALU_SWIZ")) {
				printf("swizzing r\n");
				r *= 010101010101L;
			}

			l &= mask36;					// clamp -ve sign at bit 35
			r &= mask36;

			show = 1;
			alu(opcode, l, r, &flags);
			break;

		default:							// regression tests
			test_stuff();
	}

	return 0;
}
